import copy
import re
import string
import sys
import warnings

import numpy as np

from pyfits.util import _str_to_num, _is_int, deprecated, maketrans, translate
from pyfits.verify import _Verify, _ErrList, VerifyError


__all__ = ['Card', 'CardList', 'RecordValuedKeywordCard', 'create_card',
           'create_card_from_string', 'upper_key', 'createCard',
           'createCardFromString', 'upperKey', 'Undefined']


FIX_FP_TABLE = maketrans('de', 'DE')
FIX_FP_TABLE2 = maketrans('dD', 'eE')


class Undefined:
    """Undefined value."""

    def __init__(self):
        # This __init__ is required to be here for Sphinx documentation
        pass
UNDEFINED = Undefined()


class Card(_Verify):

    # string length of a card
    length = 80

    # String for a FITS standard compliant (FSC) keyword.
    _keywd_FSC = r'[A-Z0-9_-]* *$'
    _keywd_FSC_RE = re.compile(_keywd_FSC)

    # A number sub-string, either an integer or a float in fixed or
    # scientific notation.  One for FSC and one for non-FSC (NFSC) format:
    # NFSC allows lower case of DE for exponent, allows space between sign,
    # digits, exponent sign, and exponents
    _digits_FSC = r'(\.\d+|\d+(\.\d*)?)([DE][+-]?\d+)?'
    _digits_NFSC = r'(\.\d+|\d+(\.\d*)?) *([deDE] *[+-]? *\d+)?'
    _numr_FSC = r'[+-]?' + _digits_FSC
    _numr_NFSC = r'[+-]? *' + _digits_NFSC

    # This regex helps delete leading zeros from numbers, otherwise
    # Python might evaluate them as octal values.
    _number_FSC_RE = re.compile(r'(?P<sign>[+-])?0*(?P<digt>%s)'
                                % _digits_FSC)
    _number_NFSC_RE = re.compile(r'(?P<sign>[+-])? *0*(?P<digt>%s)'
                                 % _digits_NFSC)

    # FSC commentary card string which must contain printable ASCII characters.
    _ascii_text = r'[ -~]*$'
    _comment_FSC_RE = re.compile(_ascii_text)

    # Checks for a valid value/comment string.  It returns a match object
    # for a valid value/comment string.
    # The valu group will return a match if a FITS string, boolean,
    # number, or complex value is found, otherwise it will return
    # None, meaning the keyword is undefined.  The comment field will
    # return a match if the comment separator is found, though the
    # comment maybe an empty string.
    _value_FSC_RE = re.compile(
        r'(?P<valu_field> *'
            r'(?P<valu>'

                #  The <strg> regex is not correct for all cases, but
                #  it comes pretty darn close.  It appears to find the
                #  end of a string rather well, but will accept
                #  strings with an odd number of single quotes,
                #  instead of issuing an error.  The FITS standard
                #  appears vague on this issue and only states that a
                #  string should not end with two single quotes,
                #  whereas it should not end with an even number of
                #  quotes to be precise.
                #
                #  Note that a non-greedy match is done for a string,
                #  since a greedy match will find a single-quote after
                #  the comment separator resulting in an incorrect
                #  match.
                r'\'(?P<strg>([ -~]+?|\'\'|)) *?\'(?=$|/| )|'
                r'(?P<bool>[FT])|'
                r'(?P<numr>' + _numr_FSC + r')|'
                r'(?P<cplx>\( *'
                    r'(?P<real>' + _numr_FSC + r') *, *'
                    r'(?P<imag>' + _numr_FSC + r') *\))'
            r')? *)'
        r'(?P<comm_field>'
            r'(?P<sepr>/ *)'
            r'(?P<comm>[!-~][ -~]*)?'
        r')?$')

    _value_NFSC_RE = re.compile(
        r'(?P<valu_field> *'
            r'(?P<valu>'
                r'\'(?P<strg>([ -~]+?|\'\'|)) *?\'(?=$|/| )|'
                r'(?P<bool>[FT])|'
                r'(?P<numr>' + _numr_NFSC + r')|'
                r'(?P<cplx>\( *'
                    r'(?P<real>' + _numr_NFSC + r') *, *'
                    r'(?P<imag>' + _numr_NFSC + r') *\))'
            r')? *)'
        r'(?P<comm_field>'
            r'(?P<sepr>/ *)'
            r'(?P<comm>(.|\n)*)'
        r')?$')

    # keys of commentary cards
    _commentary_keys = ['', 'COMMENT', 'HISTORY']

    def __new__(cls, key='', value='', comment=''):
        """
        Return the appropriate Card subclass depending on they key and/or
        value.
        """

        klass = cls
        val_str = _value_to_string(value)
        if len(val_str) > (Card.length - 10):
            klass = _ContinueCard
        elif isinstance(key, basestring) and key.upper().strip() == 'HIERARCH':
            klass = _HierarchCard
        return super(Card, cls).__new__(klass)

    def __init__(self, key='', value='', comment=''):
        """
        Construct a card from `key`, `value`, and (optionally) `comment`.
        Any specifed arguments, except defaults, must be compliant to FITS
        standard.

        Parameters
        ----------
        key : str, optional
            keyword name

        value : str, optional
            keyword value

        comment : str, optional
            comment
        """

        if key != '' or value != '' or comment != '':
            self.key = key
            self._update_value(value)
            self._update_comment(comment)
            # for commentary cards, value can only be strings and there
            # is no comment
            if self.key in Card._commentary_keys:
                if not isinstance(self.value, basestring):
                    raise ValueError('Value in a commentary card must be a '
                                     'string.')
        else:
            self._cardimage = ' ' * 80
        self._modified = False

    def __str__(self):
        return self.cardimage

    def _getkey(self):
        """Returns the keyword name parsed from the card image."""

        if not hasattr(self, '_key'):
            head = self._get_key_string()
            self._key = head.strip().upper()
        return self._key

    def _setkey(self, val):
        """Set the key attribute; once set it cannot be modified."""

        if hasattr(self, '_key'):
            raise AttributeError('Keyword name cannot be modified.')

        if isinstance(val, basestring):
            val = val.strip()
            if len(val) <= 8:
                val = val.upper()
                if val == 'END':
                    raise ValueError("Keyword 'END' not allowed.")
                self._check_key(val)
            else:
                if val[:8].upper() == 'HIERARCH':
                    val = val[8:].strip()
                    self.__class__ = _HierarchCard
                else:
                    raise ValueError('Keyword name %s is too long (> 8), use HIERARCH.'
                                     % val)
        else:
            raise ValueError('Keyword name %s is not a string.' % repr(val))
        self._key = val
        self._modified = True

    # TODO: It would be nice to eventually use property.getter/setter/deleter,
    # but those are not available prior to Python 2.6
    key = property(_getkey, _setkey, doc='Card keyword')

    def _getvalue(self):
        """Get the value attribute from the card image if not already set."""

        if not hasattr(self, '_value'):
            # self._value assigned directly to avoid the checks in
            # _setvalue() (and to avoid setting _value_modified = True)
            self._value = self._extract_value()
        return self._value

    def _setvalue(self, val):
        self._update_value(val)
        if hasattr(self, '_cardimage'):
            # Make sure these are saved from the old cardimage before deleting
            # it
            self.key
            self.comment
            del self._cardimage

    value = property(_getvalue, _setvalue, doc='Card value')

    def _update_value(self, val):
        """Set the value attribute."""

        if isinstance(val, (str, int, long, float, complex, bool, Undefined,
                            np.floating, np.integer, np.complexfloating,
                            np.bool_)):
            if isinstance(val, str):
                self._check_text(val)

            if (isinstance(val, float) and
                    (np.isnan(val) or np.isinf(val))):
                raise ValueError(
                    "Floating point %r values are not allowed in FITS "
                    "headers." % val)
            if not hasattr(self, '_value') or self._value != val:
                self._modified = True
            self._value_modified = True
            self._value = val
        else:
            raise ValueError('Illegal value %s.' % repr(val))

    def _getcomment(self):
        """Get the comment attribute from the card image if not already set."""

        if not hasattr(self, '_comment'):
            self._comment = self._extract_comment()
        return self._comment

    def _setcomment(self, val):
        self._update_comment(val)
        if hasattr(self, '_cardimage'):
            # Make sure these are saved from the old cardimage before deleting
            # it
            self.key
            self.value
            del self._cardimage

    comment = property(_getcomment, _setcomment, doc='Card comment')

    def _update_comment(self, val):
        """Set the comment attribute."""

        if isinstance(val, basestring):
            self._check_text(val)
        else:
            if val is not None:
                raise ValueError('Comment %s is not a string.' % repr(val))
        if not hasattr(self, '_comment') or self._comment != val:
            self._modified = True
        self._comment = val

    @property
    def cardimage(self):
        if not hasattr(self, '_cardimage'):
            self._ascardimage()
        return self._cardimage

    def _update_cardimage(self):
        """This used to do what _generate_cardimage() does, plus setting te
        value of self._cardimage.  Now it's kept mostly for symmetry.
        """

        self._cardimage = self._generate_cardimage()

    def _generate_cardimage(self):
        """Generate a (new) card image from the attributes: `key`, `value`,
        and `comment`.  Core code for `ascardimage`.
        """

        key_str = self._format_key()
        val_str = self._format_value()
        is_commentary = key_str.strip() in Card._commentary_keys
        if is_commentary:
            comment_str = ''
        else:
            comment_str = self._format_comment()

        # equal sign string
        eq_str = '= '
        if is_commentary:  # not using self.key
            eq_str = ''
            if hasattr(self, '_value'):
                # Value must be a string in commentary cards
                val_str = str(self.value)

        # put all parts together
        output = ''.join([key_str, eq_str, val_str, comment_str])

        # need this in case card-with-continue's value is shortened
        if not isinstance(self, _HierarchCard) and \
           not isinstance(self, RecordValuedKeywordCard):
            self.__class__ = Card
        else:
            key_val_len = len(key_str) + len(eq_str) + len(val_str)
            if key_val_len > Card.length:
                if isinstance(self, _HierarchCard) and \
                   key_val_len == Card.length + 1 and \
                   key_str[-1] == ' ':
                    output = ''.join([key_str[:-1], eq_str, val_str,
                                      comment_str])
                else:
                    raise ValueError('The keyword %s with its value is too '
                                     'long.' % self.key)

        if len(output) <= Card.length:
            output = '%-80s' % output

        # longstring case (CONTINUE card)
        else:
            # try not to use CONTINUE if the string value can fit in one line.
            # Instead, just truncate the comment
            if isinstance(self.value, str) and \
               len(val_str) > (Card.length - 10):
                self.__class__ = _ContinueCard
                output = self._breakup_strings()
            else:
                warnings.warn('Card is too long, comment is truncated.')
                output = output[:Card.length]

        return output

    def _format_key(self):
        # keyword string (check both _key and _cardimage attributes to avoid an
        # infinite loop
        if hasattr(self, '_key') or hasattr(self, '_cardimage'):
            return '%-8s' % self.key
        else:
            return ' ' * 8

    def _format_value(self):
        # value string
        if not (hasattr(self, '_value') or hasattr(self, '_cardimage')):
            return ''
        elif (not hasattr(self, '_value_modified') or
              not self._value_modified) and \
             isinstance(self.value, (float, np.floating, complex,
                                     np.complexfloating)):
            # Keep the existing formatting for float/complex numbers
            return '%20s' % self._valuestring
        else:
            return _value_to_string(self.value)

    def _format_comment(self):
        # comment string
        if hasattr(self, '_comment') or hasattr(self, '_cardimage'):
            if not self.comment:
                return ''
            else:
                return ' / ' + self._comment
        else:
            return ''

    # TODO: Wouldn't 'verification' be a better name for the 'option' keyword
    # argument?  'option' is pretty vague.
    @deprecated(alternative='the .cardimage attribute')
    def ascardimage(self, option='silentfix'):
        return self._ascardimage(option)

    def _ascardimage(self, option='silentfix'):
        """
        Generate a (new) card image from the attributes: `key`, `value`,
        and `comment`, or from raw string.

        Parameters
        ----------
        option : str
            Output verification option.  Must be one of ``"fix"``,
            ``"silentfix"``, ``"ignore"``, ``"warn"``, or
            ``"exception"``.  See :ref:`verify` for more info.
        """

        # Only if the card image already exist (to avoid infinite loop),
        # fix it first.
        if hasattr(self, '_cardimage'):
            self._check(option)
        self._update_cardimage()
        return self._cardimage

    @classmethod
    def fromstring(cls, cardimage):
        """
        Construct a `Card` object from a (raw) string. It will pad the
        string if it is not the length of a card image (80 columns).
        If the card image is longer than 80 columns, assume it
        contains ``CONTINUE`` card(s).
        """

        cardimage = _pad(cardimage)

        if cardimage[:8].upper() == 'HIERARCH':
            card = _HierarchCard()
        # for card image longer than 80, assume it contains CONTINUE card(s).
        elif len(cardimage) > Card.length:
            card = _ContinueCard()
        else:
            card = cls()

        card._cardimage = cardimage
        return card

    def _check_text(self, val):
        """Verify `val` to be printable ASCII text."""

        if Card._comment_FSC_RE.match(val) is None:
            self._err_text = 'Unprintable string %s' % repr(val)
            self._fixable = False
            raise ValueError(self._err_text)

    def _check_key(self, key):
        """Verify the keyword `key` to be FITS standard."""

        # use repr (not str) in case of control character
        if not Card._keywd_FSC_RE.match(key):
            self._err_text = 'Illegal keyword name %s' % repr(key)
            self._fixable = False
            raise ValueError(self._err_text)

    def _extract_value(self):
        """Extract the keyword value from the card image."""

        # for commentary cards, no need to parse further
        if self.key in Card._commentary_keys:
            return self.cardimage[8:].rstrip()

        valu = self._check(option='parse')

        if valu is None:
            raise ValueError("Unparsable card (%s), fix it first with "
                             ".verify('fix')." % self.key)
        if valu.group('bool') is not None:
            _val = valu.group('bool') == 'T'
        elif valu.group('strg') is not None:
            _val = re.sub("''", "'", valu.group('strg'))
        elif valu.group('numr') is not None:
            #  Check for numbers with leading 0s.
            numr = Card._number_NFSC_RE.match(valu.group('numr'))
            _digt = translate(numr.group('digt'), FIX_FP_TABLE2, ' ')
            if numr.group('sign') is None:
                _sign = ''
            else:
                _sign = numr.group('sign')
            _val = _str_to_num(_sign + _digt)

        elif valu.group('cplx') is not None:
            #  Check for numbers with leading 0s.
            real = Card._number_NFSC_RE.match(valu.group('real'))
            _rdigt = translate(real.group('digt'), FIX_FP_TABLE2, ' ')
            if real.group('sign') is None:
                _rsign = ''
            else:
                _rsign = real.group('sign')
            _val = _str_to_num(_rsign + _rdigt)
            imag  = Card._number_NFSC_RE.match(valu.group('imag'))
            _idigt = translate(imag.group('digt'), FIX_FP_TABLE2, ' ')
            if imag.group('sign') is None:
                _isign = ''
            else:
                _isign = imag.group('sign')
            _val += _str_to_num(_isign + _idigt)*1j
        else:
            _val = UNDEFINED

        if not hasattr(self, '_valuestring'):
            self._valuestring = valu.group('valu')
        if not hasattr(self, '_value_modified'):
            self._value_modified = False
        return _val

    def _extract_comment(self):
        """Extract the keyword value from the card image."""

        # for commentary cards, no need to parse further
        if self.key in Card._commentary_keys:
            return ''

        valu = self._check(option='parse')
        if valu is not None:
            comm = valu.group('comm')
            if isinstance(comm, str):
                return  comm.rstrip()
        return ''

    def _fix_value(self, input):
        """Fix the card image for fixable non-standard compliance."""

        _val_str = None

        # for the unparsable case
        if input is None:
            _tmp = self._get_value_comment_string()
            try:
                slash_loc = _tmp.index("/")
                self._value = _tmp[:slash_loc].strip()
                self._comment = _tmp[slash_loc + 1:].strip()
            except:
                self._value = _tmp.strip()

        elif input.group('numr') is not None:
            numr = Card._number_NFSC_RE.match(input.group('numr'))
            _val_str = translate(numr.group('digt'), FIX_FP_TABLE, ' ')
            if numr.group('sign') is not None:
                _val_str = numr.group('sign')+_val_str

        elif input.group('cplx') is not None:
            real  = Card._number_NFSC_RE.match(input.group('real'))
            _realStr = translate(real.group('digt'), FIX_FP_TABLE, ' ')
            if real.group('sign') is not None:
                _realStr = real.group('sign')+_realStr

            imag  = Card._number_NFSC_RE.match(input.group('imag'))
            _imagStr = translate(imag.group('digt'), FIX_FP_TABLE, ' ')
            if imag.group('sign') is not None:
                _imagStr = imag.group('sign') + _imagStr
            _val_str = '(%s, %s)' % (_realStr, _imagStr)

        self._valuestring = _val_str
        self._update_cardimage()

    def _locate_eq(self):
        """
        Locate the equal sign in the card image before column 10 and
        return its location.  It returns `None` if equal sign is not
        present, or it is a commentary card.
        """

        # no equal sign for commentary cards (i.e. part of the string value)
        _key = self.cardimage[:8].strip().upper()
        if _key in Card._commentary_keys:
            eq_loc = None
        else:
            if _key == 'HIERARCH':
                _limit = Card.length
            else:
                _limit = 10
            try:
                eq_loc = self.cardimage[:_limit].index("=")
            except:
                eq_loc = None
        return eq_loc

    def _get_key_string(self):
        """
        Locate the equal sign in the card image and return the string
        before the equal sign.  If there is no equal sign, return the
        string before column 9.
        """

        eq_loc = self._locate_eq()
        if eq_loc is None:
            eq_loc = 8
        start = 0
        if self.cardimage[:8].upper() == 'HIERARCH':
            start = 8
            self.__class__ = _HierarchCard
        return self.cardimage[start:eq_loc]

    def _get_value_comment_string(self):
        """
        Locate the equal sign in the card image and return the string
        after the equal sign.  If there is no equal sign, return the
        string after column 8.
        """

        eq_loc = self._locate_eq()
        if eq_loc is None:
            eq_loc = 7
        return self.cardimage[eq_loc+1:]

    def _check(self, option='ignore'):
        """Verify the card image with the specified option."""

        self._err_text = ''
        self._fix_text = ''
        self._fixable = True

        if option == 'ignore':
            return
        elif option == 'parse':

            # check the value only, no need to check key and comment for 'parse'
            result = Card._value_NFSC_RE.match(self._get_value_comment_string())

            # if not parsable (i.e. everything else) result = None
            return result
        else:

            # verify the equal sign position
            if self.key not in Card._commentary_keys and \
               self.cardimage.find('=') != 8:
                if option in ['exception', 'warn']:
                    self._err_text = \
                        'Card image is not FITS standard (equal sign not at ' \
                        'column 8).'
                    raise ValueError(self._err_text + '\n%s' % self.cardimage)
                elif option in ['fix', 'silentfix']:
                    result = self._check('parse')
                    self._fix_value(result)
                    if option == 'fix':
                        self._fix_text = \
                            'Fixed card to be FITS standard.: %s' % self.key

            # verify the key, it is never fixable
            # always fix silently the case where "=" is before column 9,
            # since there is no way to communicate back to the _keys.
            self._check_key(self.key)

            # verify the value, it may be fixable
            value = self._get_value_comment_string()
            result = Card._value_FSC_RE.match(value)
            if result is not None or self.key in Card._commentary_keys:
                return result
            else:
                if option in ['fix', 'silentfix']:
                    result = self._check('parse')
                    self._fix_value(result)
                    if option == 'fix':
                        self._fix_text = \
                            'Fixed card to be FITS standard.: %s' % self.key
                else:
                    self._err_text = \
                        'Card image is not FITS standard (unparsable value ' \
                        'string: %s).' % value
                    raise ValueError(self._err_text + '\n%s' % self.cardimage)

            # verify the comment (string), it is never fixable
            if result is not None:
                comm = result.group('comm')
                if comm is not None:
                    self._check_text(comm)

    def _ncards(self):
        return len(self.cardimage) // Card.length

    def _verify(self, option='warn'):
        """Card class verification method."""

        err = _ErrList([])
        try:
            self._check(option)
        except ValueError:
            # Trapping the ValueError raised by _check method.  Want execution to continue while printing
            # exception message.
            pass
        err.append(self.run_option(option, err_text=self._err_text,
                   fix_text=self._fix_text, fixable=self._fixable))

        return err


class RecordValuedKeywordCard(Card):
    """
    Class to manage record-valued keyword cards as described in the
    FITS WCS Paper IV proposal for representing a more general
    distortion model.

    Record-valued keyword cards are string-valued cards where the
    string is interpreted as a definition giving a record field name,
    and its floating point value.  In a FITS header they have the
    following syntax::

        keyword = 'field-specifier: float'

    where `keyword` is a standard eight-character FITS keyword name,
    `float` is the standard FITS ASCII representation of a floating
    point number, and these are separated by a colon followed by a
    single blank.  The grammar for field-specifier is::

        field-specifier:
            field
            field-specifier.field

        field:
            identifier
            identifier.index

    where `identifier` is a sequence of letters (upper or lower case),
    underscores, and digits of which the first character must not be a
    digit, and `index` is a sequence of digits.  No blank characters
    may occur in the field-specifier.  The `index` is provided
    primarily for defining array elements though it need not be used
    for that purpose.

    Multiple record-valued keywords of the same name but differing
    values may be present in a FITS header.  The field-specifier may
    be viewed as part of the keyword name.

    Some examples follow::

        DP1     = 'NAXIS: 2'
        DP1     = 'AXIS.1: 1'
        DP1     = 'AXIS.2: 2'
        DP1     = 'NAUX: 2'
        DP1     = 'AUX.1.COEFF.0: 0'
        DP1     = 'AUX.1.POWER.0: 1'
        DP1     = 'AUX.1.COEFF.1: 0.00048828125'
        DP1     = 'AUX.1.POWER.1: 1'
    """

    #
    # A group of class level regular expression definitions that allow the
    # extraction of the key, field-specifier, value, and comment from a
    # card string.
    #
    identifier = r'[a-zA-Z_]\w*'
    field = identifier + r'(\.\d+)?'
    field_specifier_s = r'%s(\.%s)*' % (field, field)
    field_specifier_val = r'(?P<keyword>%s): (?P<val>%s\s*)' \
                          % (field_specifier_s, Card._numr_FSC)
    field_specifier_NFSC_val = r'(?P<keyword>%s): (?P<val>%s\s*)' \
                               % (field_specifier_s, Card._numr_NFSC)
    keyword_val = r'\'%s\'' % field_specifier_val
    keyword_NFSC_val = r'\'%s\'' % field_specifier_NFSC_val
    keyword_val_comm = r' +%s *(/ *(?P<comm>[ -~]*))?$' % keyword_val
    keyword_NFSC_val_comm = r' +%s *(/ *(?P<comm>[ -~]*))?$' % keyword_NFSC_val
    #
    # regular expression to extract the field specifier and value from
    # a card image (ex. 'AXIS.1: 2'), the value may not be FITS Standard
    # Compliant
    #
    field_specifier_NFSC_image_RE = re.compile(field_specifier_NFSC_val)
    #
    # regular expression to extract the field specifier and value from
    # a card value; the value may not be FITS Standard Compliant
    # (ex. 'AXIS.1: 2.0e5')
    #
    field_specifier_NFSC_val_RE = re.compile(field_specifier_NFSC_val + r'$')
    #
    # regular expression to extract the key and the field specifier from a
    # string that is being used to index into a card list that contains
    # record value keyword cards (ex. 'DP1.AXIS.1')
    #
    keyword_name_RE = re.compile(r'(?P<key>%s)\.(?P<field_spec>%s)$'
                                 % (identifier, field_specifier_s))
    #
    # regular expression to extract the field specifier and value and comment
    # from the string value of a record value keyword card
    # (ex "'AXIS.1: 1' / a comment")
    #
    keyword_val_comm_RE = re.compile(keyword_val_comm)
    #
    # regular expression to extract the field specifier and value and comment
    # from the string value of a record value keyword card  that is not FITS
    # Standard Complient (ex "'AXIS.1: 1.0d12' / a comment")
    #
    keyword_NFSC_val_comm_RE = re.compile(keyword_NFSC_val_comm)

    def __init__(self, key='', value='', comment=''):
        """
        Parameters
        ----------
        key : str, optional
            The key, either the simple key or one that contains
            a field-specifier

        value : str, optional
            The value, either a simple value or one that contains a
            field-specifier

        comment : str, optional
            The comment
        """

        mo = self.keyword_name_RE.match(key)

        if mo:
            self._field_specifier = mo.group('field_spec')
            key = mo.group('key')
        else:
            if isinstance(value, str):
                if value:
                    mo = self.field_specifier_NFSC_val_RE.match(value)

                    if mo:
                        self._field_specifier = mo.group('keyword')
                        value = mo.group('val')
                        # The value should be a float, though we don't coerce
                        # ints into floats.  Anything else should be a value
                        # error
                        try:
                            value = int(value)
                        except ValueError:
                            try:
                                value = float(value)
                            except ValueError:
                                raise ValueError(
                                    "Record-valued keyword card value must be "
                                    "a floating point or integer value.")
                    else:
                        raise ValueError(
                            "Value %s must be in the form "
                            "field_specifier: value (ex. 'NAXIS: 2')" % value)
            else:
                raise ValueError('value %s is not a string' % value)

        super(RecordValuedKeywordCard, self).__init__(key, value, comment)

    @property
    def field_specifier(self):
       self._extract_value()
       return self._field_specifier

    @property
    def raw(self):
        """
        Return this card as a normal Card object not parsed as a record-valued
        keyword card.  Note that this returns a copy, so that modifications to
        it do not update the original record-valued keyword card.
        """

        key = super(RecordValuedKeywordCard, self)._getkey()
        return Card(key, self.strvalue(), self.comment)

    def _getkey(self):
        key = super(RecordValuedKeywordCard, self)._getkey()
        if not hasattr(self, '_field_specifier'):
            return key
        return '%s.%s' % (key, self._field_specifier)

    key = property(_getkey, Card.key.fset, doc=Card.key.__doc__)

    def _getvalue(self):
        """The RVKC value should always be returned as a float."""

        return float(super(RecordValuedKeywordCard, self)._getvalue())

    def _setvalue(self, val):
        if not isinstance(val, float):
            try:
                val = int(val)
            except ValueError:
                try:
                    val = float(val)
                except:
                    raise ValueError('value %s is not a float' % val)
        super(RecordValuedKeywordCard, self)._setvalue(val)

    value = property(_getvalue, _setvalue, doc=Card.value.__doc__)

    #
    # class method definitins
    #

    @classmethod
    def coerce(cls, card):
        """
        Coerces an input `Card` object to a `RecordValuedKeywordCard`
        object if the value of the card meets the requirements of this
        type of card.

        Parameters
        ----------
        card : `Card` object
            A `Card` object to coerce

        Returns
        -------
        card
            - If the input card is coercible:

                a new `RecordValuedKeywordCard` constructed from the
                `key`, `value`, and `comment` of the input card.

            - If the input card is not coercible:

                the input card
        """

        mo = cls.field_specifier_NFSC_val_RE.match(card.value)
        if mo:
            return cls(card.key, card.value, card.comment)
        else:
            return card

    @classmethod
    def upper_key(cls, key):
        """
        `classmethod` to convert a keyword value that may contain a
        field-specifier to uppercase.  The effect is to raise the
        key to uppercase and leave the field specifier in its original
        case.

        Parameters
        ----------
        key : int or str
            A keyword value that could be an integer, a key, or a
            `key.field-specifier` value

        Returns
        -------
        Integer input
            the original integer key

        String input
            the converted string
        """

        if _is_int(key):
            return key

        mo = cls.keyword_name_RE.match(key)

        if mo:
            return mo.group('key').strip().upper() + '.' + \
                   mo.group('field_spec')
        else:
            return key.strip().upper()
    # For API backwards-compatibility
    upperKey = \
        deprecated(name='upperKey', alternative='upper_key()')(upper_key)

    @classmethod
    def valid_key_value(cls, key, value=0):
        """
        Determine if the input key and value can be used to form a
        valid `RecordValuedKeywordCard` object.  The `key` parameter
        may contain the key only or both the key and field-specifier.
        The `value` may be the value only or the field-specifier and
        the value together.  The `value` parameter is optional, in
        which case the `key` parameter must contain both the key and
        the field specifier.

        Parameters
        ----------
        key : str
            The key to parse

        value : str or float-like, optional
            The value to parse

        Returns
        -------
        valid input : A list containing the key, field-specifier, value

        invalid input : An empty list

        Examples
        --------

        >>> valid_key_value('DP1','AXIS.1: 2')
        >>> valid_key_value('DP1.AXIS.1', 2)
        >>> valid_key_value('DP1.AXIS.1')
        """

        rtnKey = rtnFieldSpec = rtnValue = ''
        myKey = cls.upper_key(key)

        if isinstance(myKey, basestring):
            validKey = cls.keyword_name_RE.match(myKey)

            if validKey:
               try:
                   rtnValue = float(value)
               except ValueError:
                   pass
               else:
                   rtnKey = validKey.group('key')
                   rtnFieldSpec = validKey.group('field_spec')
            else:
                if isinstance(value, str) and \
                Card._keywd_FSC_RE.match(myKey) and len(myKey) < 9:
                    validValue = cls.field_specifier_NFSC_val_RE.match(value)
                    if validValue:
                        rtnFieldSpec = validValue.group('keyword')
                        rtnValue = validValue.group('val')
                        rtnKey = myKey

        if rtnFieldSpec:
            return [rtnKey, rtnFieldSpec, rtnValue]
        else:
            return []
    # For API backwards-compatibility
    validKeyValue = \
        deprecated(name='validKeyValue',
                   alternative='valid_key_value()')(valid_key_value)

    @classmethod
    def create(cls, key='', value='', comment=''):
        """
        Create a card given the input `key`, `value`, and `comment`.
        If the input key and value qualify for a
        `RecordValuedKeywordCard` then that is the object created.
        Otherwise, a standard `Card` object is created.

        Parameters
        ----------
        key : str, optional
            The key

        value : str, optional
            The value

        comment : str, optional
            The comment

        Returns
        -------
        card
            Either a `RecordValuedKeywordCard` or a `Card` object.
        """

        if not cls.valid_key_value(key, value):
            # This should be just a normal card
            cls = Card

        return cls(key, value, comment)
    # For API backwards-compatibility
    createCard = deprecated(name='createCard', alternative='create()')(create)

    @classmethod
    def fromstring(cls, input):
        """
        Create a card given the `input` string.  If the `input` string
        can be parsed into a key and value that qualify for a
        `RecordValuedKeywordCard` then that is the object created.
        Otherwise, a standard `Card` object is created.

        Parameters
        ----------
        input : str
            The string representing the card

        Returns
        -------
        card
            either a `RecordValuedKeywordCard` or a `Card` object
        """

        idx1 = input.find("'") + 1
        idx2 = input.rfind("'")

        if idx2 > idx1 and idx1 >= 0 and \
           cls.valid_key_value('', value=input[idx1:idx2]):
            # This calls Card.fromstring, but with the RecordValuedKeywordClass
            # as the cls argument (causing an RVKC to be created)
            return super(RecordValuedKeywordCard, cls).fromstring(input)
        else:
            # This calls Card.fromstring directly, creating a plain Card
            # object.
            return Card.fromstring(input)

    # For API backwards-compatibility
    createCardFromString = deprecated(name='createCardFromString',
                                      alternative='fromstring()')(fromstring)

    def _update_cardimage(self):
        """
        Generate a (new) card image from the attributes: `key`, `value`,
        `field_specifier`, and `comment`.  Core code for `ascardimage`.
        """

        super(RecordValuedKeywordCard, self)._update_cardimage()
        eqloc = self._cardimage.index('=')
        slashloc = self._cardimage.find('/')

        if hasattr(self, '_value_modified') and self._value_modified:
            # Bypass the automatic coertion to float here, so that values like
            # '2' will still be rendered as '2' instead of '2.0'
            value = super(RecordValuedKeywordCard, self).value
            val_str = _value_to_string(value).strip()
        else:
            val_str = self._valuestring

        val_str = "'%s: %s'" % (self._field_specifier, val_str)
        val_str = '%-20s' % val_str

        output = self._cardimage[:eqloc+2] + val_str

        if slashloc > 0:
            output = output + self._cardimage[slashloc-1:]

        if len(output) <= Card.length:
            output = '%-80s' % output

        self._cardimage = output


    def _extract_value(self):
        """Extract the keyword value from the card image."""

        valu = self._check(option='parse')

        if valu is None:
            raise ValueError(
                "Unparsable card, fix it first with .verify('fix').")

        self._field_specifier = valu.group('keyword')

        if not hasattr(self, '_valuestring'):
            self._valuestring = valu.group('val')
        if not hasattr(self, '_value_modified'):
            self._value_modified = False

        return _str_to_num(translate(valu.group('val'), FIX_FP_TABLE2, ' '))

    def strvalue(self):
        """
        Method to extract the field specifier and value from the card
        image.  This is what is reported to the user when requesting
        the value of the `Card` using either an integer index or the
        card key without any field specifier.
        """

        mo = self.field_specifier_NFSC_image_RE.search(self.cardimage)
        return self.cardimage[mo.start():mo.end()]

    def _fix_value(self, input):
        """Fix the card image for fixable non-standard compliance."""

        _val_str = None

        if input is None:
            tmp = self._get_value_comment_string()

            try:
                slash_loc = tmp.index("/")
            except:
                slash_loc = len(tmp)

            self._err_text = 'Illegal value %s' % tmp[:slash_loc]
            self._fixable = False
            raise ValueError(self._err_text)
        else:
            self._valuestring = translate(input.group('val'), FIX_FP_TABLE,
                                          ' ')
            self._update_cardimage()

    def _format_key(self):
        if hasattr(self, '_key') or hasattr(self, '_cardimage'):
            return '%-8s' % super(RecordValuedKeywordCard, self).key
        else:
            return ' ' * 8

    def _check_key(self, key):
        """
        Verify the keyword to be FITS standard and that it matches the
        standard for record-valued keyword cards.
        """

        if '.' in key:
            keyword, field_specifier = key.split('.', 1)
        else:
            keyword, field_specifier = key, None

        super(RecordValuedKeywordCard, self)._check_key(keyword)

        if field_specifier:
            if not re.match(self.field_specifier_s, key):
                self._err_text = 'Illegal keyword name %s' % repr(key)
                # TODO: Maybe fix by treating as normal card and not RVKC?
                self._fixable = False
                raise ValueError(self._err_text)

    def _check(self, option='ignore'):
        """Verify the card image with the specified `option`."""

        self._err_text = ''
        self._fix_text = ''
        self._fixable = True

        if option == 'ignore':
            return
        elif option == 'parse':
            return self.keyword_NFSC_val_comm_RE.match(
                    self._get_value_comment_string())
        else:
            # verify the equal sign position
            if self.cardimage.find('=') != 8:
                if option in ['exception', 'warn']:
                    self._err_text = \
                        'Card image is not FITS standard (equal sign not at ' \
                        'column 8).'
                    raise ValueError(self._err_text + '\n%s' % self.cardimage)
                elif option in ['fix', 'silentfix']:
                    result = self._check('parse')
                    self._fix_value(result)

                    if option == 'fix':
                        self._fix_text = \
                           'Fixed card to be FITS standard. : %s' % self.key

            # verify the key
            self._check_key(self.key)

            # verify the value
            result = \
              self.keyword_val_comm_RE.match(self._get_value_comment_string())

            if result is not None:
                return result
            else:
                if option in ['fix', 'silentfix']:
                    result = self._check('parse')
                    self._fix_value(result)

                    if option == 'fix':
                        self._fix_text = \
                              'Fixed card to be FITS standard.: %s' % self.key
                else:
                    self._err_text = \
                        'Card image is not FITS standard (unparsable value ' \
                        'string).'
                    raise ValueError(self._err_text + '\n%s' % self.cardimage)

            # verify the comment (string), it is never fixable
            if result is not None:
                _str = result.group('comm')
                if _str is not None:
                    self._check_text(_str)


class CardList(list):
    """FITS header card list class."""

    def __init__(self, cards=[], keylist=None):
        """
        Construct the `CardList` object from a list of `Card` objects.

        Parameters
        ----------
        cards
            A list of `Card` objects.
        """

        super(CardList, self).__init__(cards)

        # if the key list is not supplied (as in reading in the FITS file),
        # it will be constructed from the card list.
        if keylist is None:
            self._keys = []
            for c in self:
                if isinstance(c, RecordValuedKeywordCard):
                    self._keys.append(c.raw.key)
                else:
                    self._keys.append(c.key)
        else:
            self._keys = keylist

        # find out how many blank cards are *directly* before the END card
        self._blanks = 0
        self.count_blanks()
        self._mod = False

    def __contains__(self, key):
        return upper_key(key) in self._keys

    def __getitem__(self, key):
        """Get a `Card` by indexing or by the keyword name."""

        if isinstance(key, slice):
            return CardList(super(CardList, self).__getitem__(key),
                            self._keys[key])
        elif isinstance(key, basestring) and self._has_filter_char(key):
            return self.filter_list(key)
        else:
            idx = self.index_of(key)
            return super(CardList, self).__getitem__(idx)

    def __setitem__(self, key, value):
        """Set a `Card` by indexing or by the keyword name."""

        if isinstance(value, Card):
            idx = self.index_of(key)

            # only set if the value is different from the old one
            if str(self[idx]) != str(value):
                super(CardList, self).__setitem__(idx, value)
                if isinstance(value, RecordValuedKeywordCard):
                    self._keys[idx] = value.raw.key.upper()
                else:
                    self._keys[idx] = value.key.upper()
                self.count_blanks()
                self._mod = True
        else:
            raise ValueError('%s is not a Card' % str(value))

    def __delitem__(self, key):
        """Delete a `Card` from the `CardList`."""

        if self._has_filter_char(key):
            cardlist = self.filter_list(key)

            if len(cardlist) == 0:
                raise KeyError("Keyword '%s' not found." % key)

            for card in cardlist:
                key = card.key
                del self[key]
        else:
            idx = self.index_of(key)
            super(CardList, self).__delitem__(idx)
            del self._keys[idx]  # update the keylist
            self.count_blanks()
            self._mod = True

    def __getslice__(self, start, end):
        return self[slice(start, end)]

    def __repr__(self):
        """Format a list of cards into a string."""

        return ''.join(map(str, self))

    def __str__(self):
        """Format a list of cards into a printable string."""
        return '\n'.join(map(str, self))

    def __deepcopy__(self, memo):
        dup = CardList([copy.deepcopy(c, memo) for c in self])
        memo[id(self)] = dup
        return dup

    def _get_mod(self):
        mod = self.__dict__.get('_mod', False)
        if not mod:
            # See if any of the cards were directly modified
            for card in self:
                if card._modified:
                    self.__dict__['_mod'] = True
                    return True
        return mod

    def _set_mod(self, value):
        self.__dict__['_mod'] = value
        # If the card list is explicitly set as 'not modified' then make sure
        # the same applies to its underlying cards
        if not value:
            for card in self:
                card._modified = False

    _mod = property(_get_mod, _set_mod,
                    doc='has this card list been modified since last write')

    def copy(self):
        """Make a (deep)copy of the `CardList`."""

        return CardList([create_card_from_string(str(c)) for c in self])

    def keys(self):
        """
        Return a list of all keywords from the `CardList`.

        Keywords include ``field_specifier`` for
        `RecordValuedKeywordCard` objects.
        """

        return [card.key for card in self]

    def values(self):
        """
        Return a list of the values of all cards in the `CardList`.

        For `RecordValuedKeywordCard` objects, the value returned is
        the floating point value, exclusive of the
        ``field_specifier``.
        """

        return [c.value for c in self]

    def insert(self, pos, card, useblanks=True):
        """
        Insert a `Card` to the `CardList`.

        Parameters
        ----------
        pos : int
            The position (index, keyword name will not be allowed) to
            insert. The new card will be inserted before it.

        card : `Card` object
            The card to be inserted.

        useblanks : bool, optional
            If `useblanks` is `True`, and if there are blank cards
            directly before ``END``, it will use this space first,
            instead of appending after these blank cards, so the total
            space will not increase.  When `useblanks` is `False`, the
            card will be appended at the end, even if there are blank
            cards in front of ``END``.
        """

        if isinstance(card, Card):
            super(CardList, self).insert(pos, card)
            if isinstance(card, RecordValuedKeywordCard):
                self._keys.insert(pos, card.raw.key.upper())
            else:
                self._keys.insert(pos, card.key.upper())  # update the keylist
            self.count_blanks()
            if useblanks:
                self._use_blanks(card._ncards())

            self.count_blanks()
            self._mod = True
        else:
            raise ValueError('%s is not a Card' % str(card))

    def append(self, card, useblanks=True, bottom=False):
        """
        Append a `Card` to the `CardList`.

        Parameters
        ----------
        card : `Card` object
            The `Card` to be appended.

        useblanks : bool, optional
            Use any *extra* blank cards?

            If `useblanks` is `True`, and if there are blank cards
            directly before ``END``, it will use this space first,
            instead of appending after these blank cards, so the total
            space will not increase.  When `useblanks` is `False`, the
            card will be appended at the end, even if there are blank
            cards in front of ``END``.

        bottom : bool, optional
           If `False` the card will be appended after the last
           non-commentary card.  If `True` the card will be appended
           after the last non-blank card.
        """

        if isinstance(card, Card):
            nc = len(self) - self._blanks
            idx = nc - 1
            if not bottom:
                # locate last non-commentary card
                for idx in range(nc - 1, -1, -1):
                    if self[idx].key not in Card._commentary_keys:
                        break

            super(CardList, self).insert(idx + 1, card)
            if isinstance(card, RecordValuedKeywordCard):
                self._keys.insert(idx + 1, card.raw.key.upper())
            else:
                self._keys.insert(idx + 1, card.key.upper())
            if useblanks:
                self._use_blanks(card._ncards())
            self.count_blanks()
            self._mod = True
        else:
            raise ValueError("%s is not a Card" % str(card))

    def index_of(self, key, backward=False):
        """
        Get the index of a keyword in the `CardList`.

        Parameters
        ----------
        key : str or int
            The keyword name (a string) or the index (an integer).

        backward : bool, optional
            When `True`, search the index from the ``END``, i.e.,
            backward.

        Returns
        -------
        index : int
            The index of the `Card` with the given keyword.
        """

        if _is_int(key):
            return key
        elif isinstance(key, basestring):
            _key = key.strip().upper()
            if _key[:8] == 'HIERARCH':
                _key = _key[8:].strip()
            keys = self._keys
            if backward:
                # We have to search backwards through they key list
                keys = reversed(keys)
            try:
                idx = keys.index(_key)
            except ValueError:
                reqkey = RecordValuedKeywordCard.valid_key_value(key)
                idx = 0

                while reqkey:
                    try:
                        jdx = keys[idx:].index(reqkey[0].upper())
                        idx += jdx
                        card = self[idx]
                        if isinstance(card, RecordValuedKeywordCard) and \
                           reqkey[1] == card.field_specifier:
                            break
                    except ValueError:
                        raise KeyError('Keyword %s not found.' % repr(key))

                    idx += 1
                else:
                    raise KeyError('Keyword %s not found.' % repr(key))

            if backward:
                idx = len(keys) - idx - 1
            return idx
        else:
            raise KeyError('Illegal key data type %s' % type(key))

    def filter_list(self, key):
        """
        Construct a `CardList` that contains references to all of the cards in
        this `CardList` that match the input key value including any special
        filter keys (``*``, ``?``, and ``...``).

        Parameters
        ----------
        key : str
            key value to filter the list with

        Returns
        -------
        cardlist :
            A `CardList` object containing references to all the
            requested cards.
        """

        out_cl = CardList()

        mykey = upper_key(key)
        re_str = mykey.replace('*', '\w*') + '$'
        re_str = re_str.replace('?', '\w')
        re_str = re_str.replace('...', '\S*')
        match_RE = re.compile(re_str)

        for card in self:
            match_str = card.key

            if match_RE.match(match_str):
                out_cl.append(card)

        return out_cl
    filterList = filter_list # For API backwards-compatibility

    def count_blanks(self):
        """
        Returns how many blank cards are *directly* before the ``END``
        card.
        """

        blank = ' ' * Card.length
        for idx in range(1, len(self)):
            if str(self[-idx]) != blank:
                self._blanks = idx - 1
                break

    def _has_filter_char(self, key):
        """
        Return `True` if the input key contains one of the special filtering
        characters (``*``, ``?``, or ...).
        """

        if isinstance(key, basestring) and \
           (key.endswith('...') or key.find('*') > 0 or key.find('?') > 0):
            return True
        else:
            return False

    def _pos_insert(self, card, before, after, useblanks=True, replace=False):
        """
        Insert a `Card` to the location specified by before or after.

        The argument `before` takes precedence over `after` if both
        specified.  They can be either a keyword name or index.
        """

        if before is None:
            insertionkey = after
        else:
            insertionkey = before

        def get_insertion_idx():
            if not (isinstance(insertionkey, int) and
                    insertionkey >= len(self)):
                idx = self.index_of(insertionkey)
            else:
                idx = insertionkey

            if before is None:
                idx += 1

            return idx

        if replace:
            old_idx = self.index_of(card.key)
            insertion_idx = get_insertion_idx()

            if insertion_idx >= len(self) and old_idx == len(self) - 1:
                return

            if before is not None:
                if old_idx == insertion_idx - 1:
                    return
            elif after is not None and old_idx == insertion_idx:
                return

            del self[old_idx]

        idx = get_insertion_idx()
        self.insert(idx, card, useblanks=useblanks)

    def _use_blanks(self, how_many):
        if self._blanks > 0:
            for idx in range(min(self._blanks, how_many)):
                del self[-1] # it also delete the keylist item


def create_card(key='', value='', comment=''):
    return RecordValuedKeywordCard.create(key, value, comment)
create_card.__doc__ = RecordValuedKeywordCard.create.__doc__
# For API backwards-compatibility
createCard = deprecated(name='createCard',
                        alternative='create_card()')(create_card)


def create_card_from_string(input):
    return RecordValuedKeywordCard.fromstring(input)
create_card_from_string.__doc__ = RecordValuedKeywordCard.fromstring.__doc__
# For API backwards-compat
createCardFromString = \
        deprecated(name='createCardFromString',
                   alternative='fromstring()')(create_card_from_string)

def upper_key(key):
    return RecordValuedKeywordCard.upper_key(key)
upper_key.__doc__ = RecordValuedKeywordCard.upper_key.__doc__
# For API backwards-compat
upperKey = deprecated(name='upperKey', alternative='upper_key()')(upper_key)


class _HierarchCard(Card):
    """
    Cards begins with ``HIERARCH`` which allows keyword name longer than 8
    characters.
    """

    def _getkey(self):
        """Returns the keyword name parsed from the card image."""

        if not hasattr(self, '_key'):
            head = self._get_key_string()
            self._key = head.strip()
        return self._key

    def _format_key(self):
        if hasattr(self, '_key') or hasattr(self, '_cardimage'):
            return 'HIERARCH %s ' % self.key
        else:
            return ' ' * 8

    def _format_value(self):
        val_str = super(_HierarchCard, self)._format_value()
        # conserve space for HIERARCH cards
        return val_str.strip()


    def _verify(self, option='warn'):
        """No verification (for now)."""

        return _ErrList([])


class _ContinueCard(Card):
    """
    Cards having more than one 80-char "physical" cards, the cards after
    the first one must start with ``CONTINUE`` and the whole card must have
    string value.
    """

    def __repr__(self):
        """Format a list of cards into a printable string."""

        output = []
        for idx in range(len(self.cardimage) // 80):
            output.append(self.cardimage[idx*80:(idx+1)*80])
        return '\n'.join(output)

    def _iter_cards(self):
        ncards = self._ncards()
        for idx in range(ncards):
            # take each 80-char card as a regular card and use its methods.
            card = Card.fromstring(self.cardimage[idx*80:(idx+1)*80])
            if idx > 0 and card.key != 'CONTINUE':
                raise ValueError('Long card image must have CONTINUE cards '
                                 'after the first card.')
            if not isinstance(card.value, str):
                raise ValueError(
                    'Cards with CONTINUE must have string value.')
            yield card

    def _extract_value(self):
        """Extract the keyword value from the card image."""

        output = []
        for card in self._iter_cards():
            val = card.value.rstrip().replace("''", "'")
            # drop the ending "&"
            if val and val[-1] == '&':
                val = val[:-1]
            output.append(val)
        return ''.join(output).rstrip()

    def _extract_comment(self):
        """Extract the comment from the card image."""

        output = []
        for card in self._iter_cards():
            comm = card.comment
            if isinstance(comm, str) and comm != '':
                output.append(comm.rstrip() + ' ')
        return ''.join(output).rstrip()

    def _breakup_strings(self):
        """
        Break up long string value/comment into ``CONTINUE`` cards.
        This is a primitive implementation: it will put the value
        string in one block and the comment string in another.  Also,
        it does not break at the blank space between words.  So it may
        not look pretty.
        """

        val_len = 67
        comm_len = 64
        output = []

        # do the value string
        valfmt = "'%-s&'"
        val = self.value.replace("'", "''")
        val_list = self._words_group(val, val_len)
        for idx in range(len(val_list)):
            if idx == 0:
                headstr = "%-8s= " % self.key
            else:
                headstr = "CONTINUE  "
            valstr = valfmt % val_list[idx]
            output.append('%-80s' % (headstr + valstr))

        # do the comment string
        if self.comment is None:
            comm = ''
        else:
            comm = self.comment
        commfmt = "%-s"
        if not comm == '':
            comm_list = self._words_group(comm, comm_len)
            for idx in comm_list:
                commstr = "CONTINUE  '&' / " + commfmt % idx
                output.append('%-80s' % commstr)

        return ''.join(output)

    def _words_group(self, input, strlen):
        """
        Split a long string into parts where each part is no longer
        than `strlen` and no word is cut into two pieces.  But if
        there is one single word which is longer than `strlen`, then
        it will be split in the middle of the word.
        """

        lst = []
        _nblanks = input.count(' ')
        nmax = max(_nblanks, len(input)//strlen+1)
        arr = np.fromstring((input + ' '), dtype=(np.bytes_, 1))

        # locations of the blanks
        blank_loc = np.nonzero(arr == np.bytes_(' '))[0]
        offset = 0
        xoffset = 0
        for idx in range(nmax):
            try:
                loc = np.nonzero(blank_loc >= strlen+offset)[0][0]
                offset = blank_loc[loc-1] + 1
                if loc == 0:
                    offset = -1
            except:
                offset = len(input)

            # check for one word longer than strlen, break in the middle
            if offset <= xoffset:
                offset = xoffset + strlen

            # collect the pieces in a list
            tmp = input[xoffset:offset]
            lst.append(tmp)
            if len(input) == offset:
                break
            xoffset = offset

        return lst


def _value_to_string(value):
    """Converts a card value to its appropriate string representation as
    defined by the FITS format.
    """

    # string value should occupies at least 8 columns, unless it is
    # a null string
    if isinstance(value, str):
        if value == '':
            return "''"
        else:
            exp_val_str = value.replace("'", "''")
            val_str = "'%-8s'" % exp_val_str
            return '%-20s' % val_str

    # must be before int checking since bool is also int
    elif isinstance(value, (bool, np.bool_)):
        return '%20s' % repr(value)[0] # T or F

    elif _is_int(value):
        return '%20d' % value

    # XXX need to consider platform dependence of the format (e.g. E-009 vs. E-09)
    elif isinstance(value, (float, np.floating)):
        return '%20s' % _float_format(value)

    elif isinstance(value, (complex, np.complexfloating)):
        val_str = '(%s, %s)' % (_float_format(value.real),
                                _float_format(value.imag))
        return '%20s' % val_str

    elif isinstance(value, Undefined):
        return ''
    else:
        return ''


def _float_format(value):
    """Format a floating number to make sure it gets the decimal point."""

    value_str = '%.16G' % value
    if '.' not in value_str and 'E' not in value_str:
        value_str += '.0'
    elif 'E' in value_str:
        # On some Windows builds of Python (and possibly other platforms?) the
        # exponent is zero-padded out to, it seems, three digits.  Normalize
        # the format to pad only to two digits.
        significand, exponent = value_str.split('E')
        if exponent[0] in ('+', '-'):
            sign = exponent[0]
            exponent = exponent[1:]
        else:
            sign = ''
        value_str = '%sE%s%02d' % (significand, sign, int(exponent))

    # Limit the value string to at most 20 characters.
    str_len = len(value_str)

    if str_len > 20:
        idx = value_str.find('E')

        if idx < 0:
            value_str = value_str[:20]
        else:
            value_str = value_str[:20-(str_len-idx)] + value_str[idx:]

    return value_str


def _pad(input):
    """Pad blank space to the input string to be multiple of 80."""

    _len = len(input)
    if _len == Card.length:
        return input
    elif _len > Card.length:
        strlen = _len % Card.length
        if strlen == 0:
            return input
        else:
            return input + ' ' * (Card.length-strlen)

    # minimum length is 80
    else:
        strlen = _len % Card.length
        return input + ' ' * (Card.length-strlen)





#This is the configuration file for the pyfits namespace.

from __future__ import division # confidence high

try:
    import pkg_resources
    __version__ = pkg_resources.get_distribution('pyfits').version
except:
    __version__ = ''

# Import the pyfits core module.
import pyfits.core
import pyfits.util
from pyfits.core import *
from pyfits.util import *

__doc__ = pyfits.core.__doc__

__all__ = pyfits.core.__all__ + pyfits.util.__all__

try:
    import stsci.tools.tester
    def test(*args,**kwds):
        stsci.tools.tester.test(modname=__name__, *args, **kwds)
except ImportError:
    pass




from __future__ import with_statement

import itertools
import sys

import numpy as np

import pyfits

from pyfits.card import _pad, Card
from pyfits.tests import PyfitsTestCase
from pyfits.tests.util import ignore_warnings

from nose.tools import assert_equal, assert_false, assert_raises, assert_true


class TestHeaderFunctions(PyfitsTestCase):
    def test_update_comment(self):
        hdul = pyfits.open(self.data('arange.fits'))
        hdul[0].header.update('FOO', 'BAR', 'BAZ')
        hdul.writeto(self.temp('test.fits'))

        hdul = pyfits.open(self.temp('test.fits'), mode='update')
        hdul[0].header.ascard['FOO'].comment = 'QUX'
        hdul.close()

        hdul = pyfits.open(self.temp('test.fits'))
        assert_equal(hdul[0].header.ascard['FOO'].comment, 'QUX')

    def test_long_commentary_card(self):
        header = pyfits.Header()
        header.update('FOO', 'BAR')
        header.update('BAZ', 'QUX')
        longval = 'ABC' * 30
        header.add_history(longval)
        header.update('FRED', 'BARNEY')
        header.add_history(longval)

        assert_equal(len(header.ascard), 7)
        assert_equal(header.ascard[2].key, 'FRED')
        assert_equal(str(header[3:5]).rstrip(),
                     'HISTORY ' + longval[:72] + '\nHISTORY ' + longval[72:])
        assert_equal(str(header[5:]).rstrip(),
                     'HISTORY ' + longval[:72] + '\nHISTORY ' + longval[72:])

        header.add_history(longval, after='FOO')
        assert_equal(len(header.ascard), 9)
        assert_equal(str(header[1:3]).rstrip(),
                     'HISTORY ' + longval[:72] + '\nHISTORY ' + longval[72:])

    def test_wildcard_slice(self):
        """Test selecting a subsection of a header via wildcard matching."""

        header = pyfits.Header()
        header.update('ABC', 0)
        header.update('DEF', 1)
        header.update('ABD', 2)
        cards = header.ascard['AB*']
        assert_equal(len(cards), 2)
        assert_equal(cards[0].value, 0)
        assert_equal(cards[1].value, 2)

    def test_assign_boolean(self):
        """
        Regression test for #123. Tests assigning Python and Numpy boolean
        values to keyword values.
        """

        fooimg = _pad('FOO     =                    T')
        barimg = _pad('BAR     =                    F')
        h = pyfits.Header()
        h.update('FOO', True)
        h.update('BAR', False)
        assert_equal(h['FOO'], True)
        assert_equal(h['BAR'], False)
        assert_equal(h.ascard['FOO'].cardimage, fooimg)
        assert_equal(h.ascard['BAR'].cardimage, barimg)

        h = pyfits.Header()
        h.update('FOO', np.bool_(True))
        h.update('BAR', np.bool_(False))
        assert_equal(h['FOO'], True)
        assert_equal(h['BAR'], False)
        assert_equal(h.ascard['FOO'].cardimage, fooimg)
        assert_equal(h.ascard['BAR'].cardimage, barimg)

        h = pyfits.Header()
        h.ascard.append(pyfits.Card.fromstring(fooimg))
        h.ascard.append(pyfits.Card.fromstring(barimg))
        assert_equal(h['FOO'], True)
        assert_equal(h['BAR'], False)
        assert_equal(h.ascard['FOO'].cardimage, fooimg)
        assert_equal(h.ascard['BAR'].cardimage, barimg)

    def test_end_in_comment(self):
        """
        Regression test for #142.  Tests a case where the comment of a card
        ends with END, and is followed by several blank cards.
        """

        data = np.arange(100).reshape((10, 10))
        hdu = pyfits.PrimaryHDU(data=data)
        hdu.header.update('TESTKW', 'Test val', 'This is the END')
        # Add a couple blanks after the END string
        hdu.header.add_blank()
        hdu.header.add_blank()
        hdu.writeto(self.temp('test.fits'))

        with pyfits.open(self.temp('test.fits'), memmap=False) as hdul:
            # memmap = False to avoid leaving open a mmap to the file when we
            # access the data--this causes problems on Windows when we try to
            # overwrite the file later
            assert_true('TESTKW' in hdul[0].header)
            assert_equal(hdul[0].header, hdu.header)
            assert_true((hdul[0].data == data).all())

        # Add blanks until the header is extended to two block sizes
        while len(hdu.header.ascard) < 36:
            hdu.header.add_blank()

        with ignore_warnings():
            hdu.writeto(self.temp('test.fits'), clobber=True)

        with pyfits.open(self.temp('test.fits')) as hdul:
            assert_true('TESTKW' in hdul[0].header)
            assert_equal(hdul[0].header, hdu.header)
            assert_true((hdul[0].data == data).all())

    def test_has_key(self):
        header = pyfits.Header([Card('A', 'B', 'C'), Card('D', 'E', 'F')])
        assert_true(header.has_key('A'))
        assert_true(header.has_key('D'))
        assert_false(header.has_key('C'))

    def test_header_iter(self):
        header = pyfits.Header([Card('A', 'B'), Card('C', 'D')])
        assert_equal(list(header), ['A', 'C'])

    def test_header_clear(self):
        header = pyfits.Header([Card('A', 'B'), Card('C', 'D')])
        header.clear()
        assert_true('A' not in header)
        assert_true('C' not in header)
        assert_equal(len(header), 0)

    def test_header_items(self):
        header = pyfits.Header([Card('A', 'B'), Card('C', 'D')])
        assert_equal(header.items(), list(header.iteritems()))

        # Add a regression test specifically for #127
        header.add_history('HISTORY 1')
        header.add_history('HISTORY 2')

        assert_equal(list(header.iteritems()),
                     [('A', 'B'), ('C', 'D'), ('HISTORY', 'HISTORY 1'),
                      ('HISTORY', 'HISTORY 2')])

        assert_equal(header.items(), list(header.iteritems()))

    if sys.version_info[:2] < (3, 0):
        # 2to3 converts the iterkeys() call here to just keys(); however,
        # Header.keys() features completely different behavior from
        # Header.iterkeys() in PyFITS 3.0.x, so this test will fail in Python
        # 3--thus it is disabled
        def test_header_iterkeys(self):
            header = pyfits.Header([Card('A', 'B'), Card('C', 'D')])
            for a, b in itertools.izip(header.iterkeys(), header):
                assert_equal(a, b)

            # Add a regression test specifically for #127
            header.add_history('HISTORY 1')
            header.add_history('HISTORY 2')

            assert_equal(list(header.iterkeys()),
                         ['A', 'C', 'HISTORY', 'HISTORY'])

            # There is a built-in incongruity between Header.keys() and
            # Header.iterkeys(), in that Header.keys() removes duplicates,
            # while Header.iterkeys() does not.  This has gone away in PyFITS
            # 3.1 but for now it should remain.

    def test_header_itervalues(self):
        header = pyfits.Header([Card('A', 'B'), Card('C', 'D')])
        for a, b in itertools.izip(header.itervalues(), ['B', 'D']):
            assert_equal(a, b)

        # Add a regression test specifically for #127
        header.add_history('HISTORY 1')
        header.add_history('HISTORY 2')

        assert_equal(list(header.itervalues()),
                     ['B', 'D', 'HISTORY 1', 'HISTORY 2'])
        assert_equal(header.values(), list(header.itervalues()))

    def test_header_keys(self):
        hdul = pyfits.open(self.data('arange.fits'))
        assert_equal(hdul[0].header.keys(),
                     ['SIMPLE', 'BITPIX', 'NAXIS', 'NAXIS1', 'NAXIS2',
                      'NAXIS3', 'EXTEND'])

    def test_header_dict_like_pop(self):
        header = pyfits.Header([Card('A', 'B'), Card('C', 'D'), Card('E', 'F'),
                                Card('G', 'H')])
        assert_raises(TypeError, header.pop, 'A', 'B', 'C')

        last = header.pop('G')
        assert_equal(last, 'H')
        assert_equal(len(header), 3)
        assert_equal(header.keys(), ['A', 'C', 'E'])

        mid = header.pop('C')
        assert_equal(mid, 'D')
        assert_equal(len(header), 2)
        assert_equal(header.keys(), ['A', 'E'])

        first = header.pop('A')
        assert_equal(first, 'B')
        assert_equal(len(header), 1)
        assert_equal(header.keys(), ['E'])

        default = header.pop('X', 'Y')
        assert_equal(default, 'Y')
        assert_equal(len(header), 1)

        assert_raises(KeyError, header.pop, 'X')

    def test_popitem(self):
        header = pyfits.Header([Card('A', 'B'), Card('C', 'D'),
                                Card('E', 'F')])
        keyword, value = header.popitem()
        assert_true(keyword not in header)
        assert_equal(len(header), 2)
        keyword, value = header.popitem()
        assert_true(keyword not in header)
        assert_equal(len(header), 1)
        keyword, value = header.popitem()
        assert_true(keyword not in header)
        assert_equal(len(header), 0)
        assert_raises(KeyError, header.popitem)

    def test_setdefault(self):
        header = pyfits.Header([Card('A', 'B'), Card('C', 'D'),
                                Card('E', 'F')])
        assert_equal(header.setdefault('A'), 'B')
        assert_equal(header.setdefault('C'), 'D')
        assert_equal(header.setdefault('E'), 'F')
        assert_equal(len(header), 3)
        assert_equal(header.setdefault('G', 'H'), 'H')
        assert_equal(len(header), 4)
        assert_true('G' in header)
        assert_equal(header.setdefault('G', 'H'), 'H')
        assert_equal(len(header), 4)

    def test_assign_inf_nan(self):
        """
        Regression test for https://github.com/spacetelescope/PyFITS/issues/11

        For the time being it should not be possible to assign the floating
        point values inf or nan to a header value, since this is not defined by
        the FITS standard.
        """

        h = pyfits.Header()

        # There is an obscure cross-platform issue that prevents things like
        # float('nan') on Windows on older versions of Python; hence it is
        # unlikely to come up in practice
        if not (sys.platform.startswith('win32') and
                sys.version_info[:2] < (2, 6)):
           assert_raises(ValueError, h.update, 'TEST', float('nan'))
           assert_raises(ValueError, h.update, 'TEST', float('inf'))

        assert_raises(ValueError, h.update, 'TEST', np.nan)
        assert_raises(ValueError, h.update, 'TEST', np.inf)


class TestRecordValuedKeywordCards(PyfitsTestCase):
    """Tests for handling of record-valued keyword cards as used by the FITS
    WCS Paper IV proposal.

    These tests are derived primarily from the release notes for PyFITS 1.4 (in
    which this feature was first introduced.
    """

    def setup(self):
        super(TestRecordValuedKeywordCards, self).setup()
        self._test_header = pyfits.Header()
        self._test_header.update('DP1', 'NAXIS: 2')
        self._test_header.update('DP1', 'AXIS.1: 1')
        self._test_header.update('DP1', 'AXIS.2: 2')
        self._test_header.update('DP1', 'NAUX: 2')
        self._test_header.update('DP1', 'AUX.1.COEFF.0: 0')
        self._test_header.update('DP1', 'AUX.1.POWER.0: 1')
        self._test_header.update('DP1', 'AUX.1.COEFF.1: 0.00048828125')
        self._test_header.update('DP1', 'AUX.1.POWER.1: 1')

    def test_field_specifier_case_senstivity(self):
        """The keyword portion of an RVKC should still be case-insensitive, but
        the the field-specifier portion should be case-sensitive.
        """

        header = pyfits.Header()
        header.update('abc.def', 1)
        header.update('abc.DEF', 2)
        assert_equal(header['abc.def'], 1)
        assert_equal(header['ABC.def'], 1)
        assert_equal(header['aBc.def'], 1)
        assert_equal(header['ABC.DEF'], 2)
        assert_false('ABC.dEf' in header)

    def test_get_rvkc_by_index(self):
        """Returning a RVKC from a header via index lookup should return the
        entire string value of the card, including the field-specifier.
        """

        assert_equal(self._test_header[0], 'NAXIS: 2')
        assert_equal(self._test_header[1], 'AXIS.1: 1')

    def test_get_rvkc_by_keyword(self):
        """Returning a RVKC just via the keyword name should return the entire
        string value of the first card with that keyword.
        """

        assert_equal(self._test_header['DP1'], 'NAXIS: 2')

    def test_get_rvkc_by_keyword_and_field_specifier(self):
        """Returning a RVKC via the full keyword/field-specifier combination
        should return the floating point value associated with the RVKC.
        """

        assert_equal(self._test_header['DP1.NAXIS'], 2.0)
        assert_true(isinstance(self._test_header['DP1.NAXIS'], float))
        assert_equal(self._test_header['DP1.AUX.1.COEFF.1'], 0.00048828125)

    def test_access_nonexistent_rvkc(self):
        """Accessing a nonexistent RVKC should raise an IndexError for
        index-based lookup, or a KeyError for keyword lookup (like a normal
        card).
        """

        assert_raises(IndexError, lambda x: self._test_header[x], 8)
        assert_raises(KeyError, lambda k: self._test_header[k], 'DP1.AXIS.3')

    def test_update_rvkc(self):
        """A RVKC can be updated either via index or keyword access."""

        self._test_header[0] = 3
        assert_equal(self._test_header['DP1.NAXIS'], 3.0)
        assert_true(isinstance(self._test_header['DP1.NAXIS'], float))
        assert_equal(self._test_header[0], 'NAXIS: 3')

        self._test_header['DP1.AXIS.1'] = 1.1
        assert_equal(self._test_header['DP1.AXIS.1'], 1.1)
        assert_equal(self._test_header[1], 'AXIS.1: 1.1')

    def test_rvkc_insert_after(self):
        """It should be possible to insert a new RVKC after an existing one
        specified by the full keyword/field-specifier combination."""

        self._test_header.update('DP1', 'AXIS.3: 1', 'a comment',
                                 after='DP1.AXIS.2')
        assert_equal(self._test_header[3], 'AXIS.3: 1')
        assert_equal(self._test_header['DP1.AXIS.3'], 1)

    def test_rvkc_delete(self):
        """Deleting a RVKC should work as with a normal card by using the full
        keyword/field-spcifier combination.
        """

        del self._test_header['DP1.AXIS.1']
        assert_equal(len(self._test_header), 7)
        assert_equal(self._test_header[0], 'NAXIS: 2')
        assert_equal(self._test_header[1], 'AXIS.2: 2')

    def test_pattern_matching_keys(self):
        """Test the keyword filter strings with RVKCs."""

        cl = self._test_header['DP1.AXIS.*']
        assert_true(isinstance(cl, pyfits.CardList))
        assert_equal(
            [l.strip() for l in str(cl).splitlines()],
            ["DP1     = 'AXIS.1: 1'",
             "DP1     = 'AXIS.2: 2'"])

        cl = self._test_header['DP1.N*']
        assert_equal(
            [l.strip() for l in str(cl).splitlines()],
            ["DP1     = 'NAXIS: 2'",
             "DP1     = 'NAUX: 2'"])

        cl = self._test_header['DP1.AUX...']
        assert_equal(
            [l.strip() for l in str(cl).splitlines()],
            ["DP1     = 'AUX.1.COEFF.0: 0'",
             "DP1     = 'AUX.1.POWER.0: 1'",
             "DP1     = 'AUX.1.COEFF.1: 0.00048828125'",
             "DP1     = 'AUX.1.POWER.1: 1'"])

        cl = self._test_header['DP?.NAXIS']
        assert_equal(
            [l.strip() for l in str(cl).splitlines()],
            ["DP1     = 'NAXIS: 2'"])

        cl = self._test_header['DP1.A*S.*']
        assert_equal(
            [l.strip() for l in str(cl).splitlines()],
            ["DP1     = 'AXIS.1: 1'",
             "DP1     = 'AXIS.2: 2'"])

    def test_pattern_matching_key_deletion(self):
        """Deletion by filter strings should work."""

        del self._test_header['DP1.A*...']
        assert_equal(len(self._test_header), 2)
        assert_equal(self._test_header[0], 'NAXIS: 2')
        assert_equal(self._test_header[1], 'NAUX: 2')

    def test_successive_pattern_matching(self):
        """A card list returned via a filter string should be further
        filterable."""

        cl = self._test_header['DP1.A*...']
        assert_equal(
            [l.strip() for l in str(cl).splitlines()],
            ["DP1     = 'AXIS.1: 1'",
             "DP1     = 'AXIS.2: 2'",
             "DP1     = 'AUX.1.COEFF.0: 0'",
             "DP1     = 'AUX.1.POWER.0: 1'",
             "DP1     = 'AUX.1.COEFF.1: 0.00048828125'",
             "DP1     = 'AUX.1.POWER.1: 1'"])

        cl2 = cl['*.*AUX...']
        assert_equal(
            [l.strip() for l in str(cl2).splitlines()],
            ["DP1     = 'AUX.1.COEFF.0: 0'",
             "DP1     = 'AUX.1.POWER.0: 1'",
             "DP1     = 'AUX.1.COEFF.1: 0.00048828125'",
             "DP1     = 'AUX.1.POWER.1: 1'"])

    def test_rvkc_in_cardlist_keys(self):
        """The CardList.keys() method should return full keyword/field-spec
        values for RVKCs.
        """

        cl = self._test_header['DP1.AXIS.*']
        assert_equal(cl.keys(), ['DP1.AXIS.1', 'DP1.AXIS.2'])

    def test_rvkc_in_cardlist_values(self):
        """The CardList.values() method should return the values of all RVKCs
        as floating point values.
        """

        cl = self._test_header['DP1.AXIS.*']
        assert_equal(cl.values(), [1.0, 2.0])

    def test_rvkc_cardlist_indexing(self):
        """RVKC should be retrievable from CardLists using standard index or
        keyword-based lookup.
        """

        cl = self._test_header['DP1.AXIS.*']
        assert_equal(str(cl[0]).strip(), "DP1     = 'AXIS.1: 1'")
        assert_equal(str(cl['DP1.AXIS.2']).strip(), "DP1     = 'AXIS.2: 2'")

    def test_rvkc_value_attribute(self):
        """Individual card values should be accessible by the .value attribute
        (which should return a float).
        """

        cl = self._test_header['DP1.AXIS.*']
        assert_equal(cl[0].value, 1.0)
        assert_true(isinstance(cl[0].value, float))

    def test_rvkc_cardlist_deletion(self):
        """Modifying RVKCs in a CardList should be reflected in any Header that
        those cards belong to, but deleting a RVKC from a CardList should not
        remove that card from the Header.
        """

        cl = self._test_header['DP1.AXIS.*']
        cl[0].value = 4.0
        assert_equal(self._test_header['DP1.AXIS.1'], 4.0)
        del cl[0]
        assert_raises(KeyError, lambda k: cl[k], 'DP1.AXIS.1')
        assert_equal(self._test_header['DP1.AXIS.1'], 4.0)

    def test_rvkc_constructor(self):
        """Test direct creation of RVKC objects."""

        c1 = pyfits.RecordValuedKeywordCard('DP1', 'NAXIS: 2',
                                            'Number of variables')
        c2 = pyfits.RecordValuedKeywordCard('DP1.AXIS.1', 1.0, 'Axis number')

        assert_equal(c1.key, 'DP1.NAXIS')
        assert_equal(c1.value, 2.0)
        assert_equal(c1.comment, 'Number of variables')
        assert_equal(c1.field_specifier, 'NAXIS')
        assert_equal(c2.key, 'DP1.AXIS.1')
        assert_equal(c2.value, 1.0)
        assert_equal(c2.comment, 'Axis number')
        assert_equal(c2.field_specifier, 'AXIS.1')

        # RVKCs created with this constructor should verify without any
        # problems
        c1.verify('exception')
        c2.verify('exception')

    def test_rvkc_fromstring(self):
        """Test creation of RVKC from their string representation."""

        c1 = pyfits.RecordValuedKeywordCard().fromstring(
            "DP1     = 'NAXIS: 2' / Number of independent variables")
        c2 = pyfits.RecordValuedKeywordCard().fromstring(
            "DP1     = 'AXIS.1: X' / Axis number")

        assert_equal(str(c1).strip(),
                     "DP1     = 'NAXIS: 2' / Number of independent variables")
        assert_equal(str(c2).strip(),
                     "DP1     = 'AXIS.1: X' / Axis number")
        # Since c2's value is wrong for a RVKC it should be a normal card
        assert_false(isinstance(c2, pyfits.RecordValuedKeywordCard))



from __future__ import division

import functools
import itertools
import mmap
import os
import platform
import signal
import sys
import tempfile
import threading
import warnings

import numpy as np


__all__ = ['Extendable', 'register_extension', 'register_extensions',
           'unregister_extensions']


BLOCK_SIZE = 2880  # the FITS block size


# TODO: I'm somewhat of the opinion that this should go in pyfits.core, but for
# now that would create too much complication with imports (as many modules
# need to use this).  Eventually all the intra-package imports will be removed
# from pyfits.core, simplifying matters.  But for now they remain for
# backwards-compatibility.
class Extendable(type):
    _extensions = {}

    def __call__(cls, *args, **kwargs):
        if cls in cls._extensions:
            cls = cls._extensions[cls]
        self = cls.__new__(cls, *args, **kwargs)
        self.__init__(*args, **kwargs)
        return self

# TODO: Fix this in Python3--commented out in the meantime
#    def __getattribute__(cls, attr):
#        orig_cls = cls
#        if attr != '_extensions' and cls in cls._extensions:
#            cls = cls._extensions[cls]
#        return super(Extendable, cls).__getattribute__(attr)

    @classmethod
    def register_extension(cls, extension, extends=None, silent=False):
        """
        Register an extension class.  This class will be used in all future
        instances of the class it extends.

        By default, the class it extends
        will automatically be its immediate superclass.  In
        multiple-inheritence cases, the left-most superclass is used.  In other
        words, the first class in the extension's MRO (after itself).

        Parameters
        ----------
        extension : class
            The class object of a superclass of an Extendable class (that is,
            any class with the `Extendable` metaclass

        extends : class, optional
            Override the default behavior of extending the immediate
            superclass.  Either a single class to extend may be specified, or
            an iterable of classes.  The classes to extend must still have the
            `Extendable` metaclass.

        silent : bool, optional
            If `True`, does not output any warnings
        """

        # If the extension class itself is not Extendable then we know it's no
        # good, since it has to be a sublcass of an Extendable.
        if not isinstance(extension, Extendable):
            raise TypeError("Class '%s' is not a subclass of an Extendable "
                            "class." % extension.__name__)

        if extends:
            try:
                extends = iter(extends)
            except TypeError:
                extends = [extends]
        else:
            extends = [extension.mro()[1]]

        # Don't allow the extension to be registered if *any* of the classes it
        # extends are not Extendable
        for c in extends:
            if not isinstance(c, Extendable):
                raise TypeError("Class '%s' is not an Extendable class."
                                % c.__name__)

        for c in extends:
            if c in cls._extensions:
                if not silent:
                    warnings.warn(
                        "Extension '%s' for '%s' being replaced with '%s'."
                        % (cls._extensions[c].__name__, c.__name__,
                           extension.__name__))
                # This class has already been extended by a different class, so
                # first we need to undo that
                cls._unextend_subclasses(c, extension)
            # It's imperative that _extend_subclasses is called first;
            # otherwise the extension will override c.__subclasses__!
            cls._extend_subclasses(c, extension)
            cls._extensions[c] = extension

    @classmethod
    def register_extensions(cls, extensions, silent=False):
        """
        Register multiple extensions at once from a dict mapping extensions to
        the classes they extend.
        """

        for k, v in extensions.iteritems():
            if not isinstance(k, Extendable):
                raise TypeError("Extension class '%s' is not a subclass of "
                                "an Extendable class." % k.__name__)
            if not isinstance(v, Extendable):
                raise TypeError("Class '%s' is not an Extendable class.")

        for k, v in extensions.iteritems():
            if not silent and v in cls._extensions:
                warnings.showwarning(
                    "Extension '%s' for '%s' being replaced with '%s'."
                    % (cls._extensions[v].__name__, v.__name__, k.__name__))
            cls._extend_subclasses(v, k)
            cls._extensions[v] = k

    @classmethod
    def unregister_extensions(cls, extensions):
        """
        Remove one or more extension classes from the extension registry.

        If the class is not in the registry this is silently ignored.
        """

        try:
            extensions = set(extensions)
        except TypeError:
            extensions = set([extensions])

        for k, v in cls._extensions.items():
            if v in extensions:
                del cls._extensions[k]
                cls._unextend_subclasses(k, v)

    @classmethod
    def _extend_subclasses(cls, extendable, extension):
        for s in extendable.__subclasses__():
            if s is extension:
                # Don't want the extension to have itself as a base
                continue
            bases = list(s.__bases__)
            idx = bases.index(extendable)
            bases = bases[:idx] + [extension] + bases[idx + 1:]
            s.__bases__ = tuple(bases)

    @classmethod
    def _unextend_subclasses(cls, extendable, extension):
        # Since the subclasses' bases were changed, they will now be listed as
        # subclasses of the extension
        for s in extension.__subclasses__():
            if s is extension:
                continue
            bases = list(s.__bases__)
            idx = bases.index(extension)
            bases = bases[:idx] + [extendable] + bases[idx + 1:]
            s.__bases__ = tuple(bases)

# Some shortcuts
register_extension = Extendable.register_extension
register_extensions = Extendable.register_extensions
# unregister_extension is currently the same as just unregister_extensions (it
# won't balk if you still try to pass more than one)--it's just provided for
# symmetry's sake
unregister_extension = Extendable.unregister_extensions
unregister_extensions = Extendable.unregister_extensions


def _with_extensions(func):
    """
    This decorator exists mainly to support use of the new extension system in
    functions that still have a classExtensions keyword argument (which should
    be deprecated).

    This registers the extensions passed in classExtensions and unregisters
    them when the function exits.  It should be clear that any objects that
    persist after the function exits will still use the extension classes they
    were created from.
    """

    @functools.wraps(func)
    def _with_extensions_wrapper(*args, **kwargs):
        extension_classes = []
        if 'classExtensions' in kwargs:
            extensions = kwargs['classExtensions']
            warnings.warn('The classExtensions argument is deprecated.  '
                          'Instead call pyfits.register_extensions(%s) once '
                          'before any code that uses those extensions.'
                          % repr(extensions), DeprecationWarning)
            if extensions:
                register_extensions(extensions)
                extension_classes = extensions.values()
            del kwargs['classExtensions']
        try:
            return func(*args, **kwargs)
        finally:
            if extension_classes:
                unregister_extensions(extension_classes)

    return _with_extensions_wrapper


def itersubclasses(cls, _seen=None):
    """
    itersubclasses(cls)

    Generator over all subclasses of a given class, in depth first order.

    >>> list(itersubclasses(int)) == [bool]
    True
    >>> class A(object): pass
    >>> class B(A): pass
    >>> class C(A): pass
    >>> class D(B,C): pass
    >>> class E(D): pass
    >>>
    >>> for cls in itersubclasses(A):
    ...     print(cls.__name__)
    B
    D
    E
    C
    >>> # get ALL (new-style) classes currently defined
    >>> [cls.__name__ for cls in itersubclasses(object)] #doctest: +ELLIPSIS
    ['type', ...'tuple', ...]

    From http://code.activestate.com/recipes/576949/
    """

    if not isinstance(cls, type):
        raise TypeError('itersubclasses must be called with '
                        'new-style classes, not %.100r' % cls)
    if _seen is None: _seen = set()
    try:
        subs = cls.__subclasses__()
    except TypeError: # fails only when cls is type
        subs = cls.__subclasses__(cls)
    for sub in subs:
        if sub not in _seen:
            _seen.add(sub)
            yield sub
            for sub in itersubclasses(sub, _seen):
                yield sub


class lazyproperty(object):
    """
    Works similarly to property(), but computes the value only once.

    Adapted from the recipe at
    http://code.activestate.com/recipes/363602-lazy-property-evaluation
    """

    def __init__(self, fget, fset=None, fdel=None):
        self._fget = fget
        self._fset = fset
        self._fdel = fdel

    def __get__(self, obj, owner=None):
        if obj is None:
            return self
        key = self._fget.func_name
        if key not in obj.__dict__:
            val = self._fget(obj)
            obj.__dict__[key] = val
            return val
        else:
            return obj.__dict__[key]

    def __set__(self, obj, val):
        obj_dict = obj.__dict__
        func_name = self._fget.func_name
        if self._fset:
            ret = self._fset(obj, val)
            if ret is not None and obj_dict.get(func_name) is ret:
                # By returning the value set the setter signals that it took
                # over setting the value in obj.__dict__; this mechanism allows
                # it to override the input value
                return
        obj_dict[func_name] = val

    def __delete__(self, obj):
        if self._fdel:
            self._fdel(obj)
        key = self._fget.func_name
        if key in obj.__dict__:
            del obj.__dict__[key]


def deprecated(message='', name='', alternative='', pending=False):
    """
    Used to mark a function as deprecated.

    TODO: Provide a class deprecation marker as well.

    To mark an attribute as deprecated, replace that attribute with a
    depcrecated property.

    Parameters
    ------------
    message : str, optional
        Override the default deprecation message.  The format specifier
        %(func)s may be used for the name of the function, and %(alternative)s
        may be used in the deprecation message to insert the name of an
        alternative to the deprecated function.

    name : str, optional
        The name of the deprecated function; if not provided the name is
        automatically determined from the passed in function, though this is
        useful in the case of renamed functions, where the new function is just
        assigned to the name of the deprecated function.  For example:
            def new_function():
                ...
            oldFunction = new_function

    alternative : str, optional
        An alternative function that the user may use in place of the
        deprecated function.  The deprecation warning will tell the user about
        this alternative if provided.

    pending : bool, optional
        If True, uses a PendingDeprecationWarning instead of a
        DeprecationWarning.

    """

    def deprecate(func):
        if isinstance(func, classmethod):
            try:
                func = func.__func__
            except AttributeError:
                # classmethods in Python2.6 and below lack the __func__
                # attribute so we need to hack around to get it
                method = func.__get__(None, object)
                if hasattr(method, '__func__'):
                    func = method.__func__
                elif hasattr(method, 'im_func'):
                    func = method.im_func
                else:
                    # Nothing we can do really...  just return the original
                    # classmethod
                    return func
            is_classmethod = True
        else:
            is_classmethod = False

        @functools.wraps(func)
        def deprecated_func(*args, **kwargs):
            # _message and _name are necessary; otherwise assignments to name
            # and message will cause the interpreter to treat them as local
            # variables, instead of encapuslated variables
            _message = message
            _name = name
            if not _name:
                _name = func.__name__

            if not _message or type(_message) == type(deprecate):
                if pending:
                    _message = 'The %(func)s function will be deprecated in' \
                               'a future version.'
                else:
                    _message = 'The %(func)s function is deprecated and may ' \
                               'be removed in a future version.'
                if alternative:
                    _message += '  Use %(alternative)s instead.'

            if pending:
                category = PendingDeprecationWarning
            else:
                category = DeprecationWarning

            warnings.warn(
                _message % {'func': _name, 'alternative': alternative},
                category, stacklevel=2)

            return func(*args, **kwargs)
        if is_classmethod:
            deprecated_func = classmethod(deprecated_func)
        return deprecated_func

    if type(message) == type(deprecate):
        return deprecate(message)

    return deprecate


def ignore_sigint(func):
    """
    This decorator registers a custom SIGINT handler to catch and ignore SIGINT
    until the wrapped function is completed.
    """

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        # Get the name of the current thread and determine if this is a single
        # treaded application
        curr_thread = threading.currentThread()
        single_thread = (threading.activeCount() == 1 and
                         curr_thread.getName() == 'MainThread')

        class SigintHandler(object):
            def __init__(self):
                self.sigint_received = False

            def __call__(self, signum, frame):
                warnings.warn('KeyboardInterrupt ignored until %s is '
                              'complete!' % func.__name__)
                self.sigint_received = True

        sigint_handler = SigintHandler()

        # Define new signal interput handler
        if single_thread:
            # Install new handler
            old_handler = signal.signal(signal.SIGINT, sigint_handler)

        try:
            func(*args, **kwargs)
        finally:
            if single_thread:
                if old_handler is not None:
                    signal.signal(signal.SIGINT, old_handler)
                else:
                    signal.signal(signal.SIGINT, signal.SIG_DFL)

                if sigint_handler.sigint_received:
                    raise KeyboardInterrupt

    return wrapped


def pairwise(iterable):
    """Return the items of an iterable paired with its next item.

    Ex: s -> (s0,s1), (s1,s2), (s2,s3), ....
    """

    a, b = itertools.tee(iterable)
    for _ in b:
        # Just a little trick to advance b without having to catch
        # StopIter if b happens to be empty
        break
    return itertools.izip(a, b)


def encode_ascii(s):
    """
    In Python 2 this is a no-op.  Strings are left alone.  In Python 3 this
    will be replaced with a function that actually encodes unicode strings to
    ASCII bytes.
    """

    return s


def decode_ascii(s):
    """
    In Python 2 this is a no-op.  Strings are left alone.  In Python 3 this
    will be replaced with a function that actually decodes ascii bytes to
    unicode.
    """

    return s


def isreadable(f):
    """
    Returns True if the file-like object can be read from.  This is a common-
    sense approximation of io.IOBase.readable.
    """

    if hasattr(f, 'closed') and f.closed:
        # This mimics the behavior of io.IOBase.readable
        raise ValueError('I/O operation on closed file')

    if not hasattr(f, 'read'):
        return False

    if hasattr(f, 'mode') and not any((c in f.mode for c in 'r+')):
        return False

    # Not closed, has a 'read()' method, and either has no known mode or a
    # readable mode--should be good enough to assume 'readable'
    return True


def iswritable(f):
    """
    Returns True if the file-like object can be written to.  This is a common-
    sense approximation of io.IOBase.writable.
    """

    if hasattr(f, 'closed') and f.closed:
        # This mimics the behavior of io.IOBase.writable
        raise ValueError('I/O operation on closed file')

    if not hasattr(f, 'write'):
        return False

    if hasattr(f, 'mode') and not any((c in f.mode for c in 'wa+')):
        return False

    # Note closed, has a 'write()' method, and either has no known mode or a
    # mode that supports writing--should be good enough to assume 'writable'
    return True


def isfile(f):
    """
    Returns True if the given object represents an OS-level file (that is,
    isinstance(f, file)).

    On Python 3 this also returns True if the given object is higher level
    wrapper on top of a FileIO object, such as a TextIOWrapper.
    """

    return isinstance(f, file)


def fileobj_open(filename, mode):
    """
    A wrapper around the `open()` builtin.

    This exists because in Python 3, `open()` returns an `io.BufferedReader` by
    default.  This is bad, because `io.BufferedReader` doesn't support random
    access, which we need in some cases.  In the Python 3 case (implemented in
    the py3compat module) we must call open with buffering=0 to get a raw
    random-access file reader.
    """

    return open(filename, mode)


def fileobj_name(f):
    """
    Returns the 'name' of file-like object f, if it has anything that could be
    called its name.  Otherwise f's class or type is returned.  If f is a
    string f itself is returned.
    """

    if isinstance(f, basestring):
        return f
    elif hasattr(f, 'name'):
        return f.name
    elif hasattr(f, 'filename'):
        return f.filename
    elif hasattr(f, '__class__'):
        return str(f.__class__)
    else:
        return str(type(f))


def fileobj_closed(f):
    """
    Returns True if the given file-like object is closed or if f is not a
    file-like object.
    """

    if hasattr(f, 'closed'):
        return f.closed
    elif hasattr(f, 'fileobj') and hasattr(f.fileobj, 'closed'):
        return f.fileobj.closed
    elif hasattr(f, 'fp') and hasattr(f.fp.closed):
        return f.fp.closed
    else:
        return True


def fileobj_is_binary(f):
    """
    Returns True if the give file or file-like object has a file open in binary
    mode.  When in doubt, returns True by default.
    """

    # TODO: In Python 3 it might be more reliable to check if the fileobj is a
    # text reader or a binary reader

    mode = fileobj_mode(f)
    if mode:
        return 'b' in mode
    else:
        return True


def fileobj_mode(f):
    """
    Returns the 'mode' string of a file-like object if such a thing exists.
    Otherwise returns None.
    """

    # Go from most to least specific--for example gzip objects have a 'mode'
    # attribute, but it's not analogous to the file.mode attribute
    if hasattr(f, 'fileobj') and hasattr(f.fileobj, 'mode'):
        mode = f.fileobj.mode
    elif hasattr(f, 'fp') and hasattr(f.fp, 'mode'):
        mode = f.fp.mode
    elif hasattr(f, 'mode'):
        mode = f.mode
    else:
        mode = None

    if isinstance(mode, basestring):
        # On Python 3 files opened in 'a' mode actually get opened in 'w' or
        # 'r+' instead of 'a', since the behavior of 'a' mode is not normalized
        # across systems.  So we should normalize in the same way...
        if 'a' in mode:
            if '+' in mode:
                mode = mode.replace('a', 'r')
            else:
                mode = mode.replace('a', 'w')
    return mode


def translate(s, table, deletechars):
    """
    This is a version of string/unicode.translate() that can handle string or
    unicode strings the same way using a translation table made with
    string.maketrans.
    """

    if isinstance(s, str):
        return s.translate(table, deletechars)
    elif isinstance(s, unicode):
        table = dict((x, ord(table[x])) for x in range(256)
                     if ord(table[x]) != x)
        for c in deletechars:
            table[ord(c)] = None
        return s.translate(table)


def indent(s, shift=1, width=4):
    return '\n'.join(' ' * (width * shift) + l if l else ''
                     for l in s.splitlines())


def _array_from_file(infile, dtype, count, sep):
    """Create a numpy array from a file or a file-like object."""

    if isfile(infile):
        return np.fromfile(infile, dtype=dtype, count=count, sep=sep)
    else:
        # treat as file-like object with "read" method; this includes gzip file
        # objects, because numpy.fromfile just reads the compressed bytes from
        # their underlying file object, instead of the decompresed bytes
        read_size = np.dtype(dtype).itemsize * count
        s = infile.read(read_size)
        return np.fromstring(s, dtype=dtype, count=count, sep=sep)


def _array_to_file(arr, outfile):
    """Write a numpy array to a file or a file-like object."""

    if isfile(outfile):
        def write(a, f):
            a.tofile(f)
    else:
        # treat as file-like object with "write" method
        def write(a, f):
            _write_string(f, a.tostring())

    # Implements a workaround for a bug deep in OSX's stdlib file writing
    # functions; on 64-bit OSX it is not possible to correctly write a number
    # of bytes greater than 2 ** 32 and divisble by 4096 (or possibly 8192--
    # whatever the default blocksize for the filesystem is).
    # This issue should have a workaround in Numpy too, but hasn't been
    # implemented there yet: https://github.com/astropy/astropy/issues/839
    osx_write_limit = (2 ** 32) - 1

    if (sys.platform == 'darwin' and platform.architecture()[0] == '64bit' and
            arr.nbytes >= osx_write_limit + 1 and arr.nbytes % 4096 == 0):
        idx = 0
        # chunksize is a count of elements in the array, not bytes
        chunksize = osx_write_limit // arr.itemsize
        while idx < arr.nbytes:
            write(arr[idx:idx + chunksize], outfile)
            idx += chunksize
    else:
        write(arr, outfile)


def _write_string(f, s):
    """
    Write a string to a file, encoding to ASCII if the file is open in binary
    mode, or decoding if the file is open in text mode.
    """

    # Assume if the file object doesn't have a specific mode, that the mode is
    # binary
    binmode = fileobj_is_binary(f)

    if binmode and isinstance(s, unicode):
        s = encode_ascii(s)
    elif not binmode and not isinstance(f, unicode):
        s = decode_ascii(s)
    f.write(s)


def _chunk_array(arr, CHUNK_SIZE=2 ** 25):
    """
    Yields subviews of the given array.  The number of rows is
    selected so it is as close to CHUNK_SIZE (bytes) as possible.
    """

    if len(arr) == 0:
        return
    if isinstance(arr, FITS_rec):
        arr = np.asarray(arr)
    row_size = arr[0].size
    rows_per_chunk = max(min(CHUNK_SIZE // row_size, len(arr)), 1)
    for idx in range(0, len(arr), rows_per_chunk):
        yield arr[idx:idx + rows_per_chunk,...]


def _convert_array(array, dtype):
    """
    Converts an array to a new dtype--if the itemsize of the new dtype is
    the same as the old dtype and both types are not numeric, a view is
    returned.  Otherwise a new array must be created.
    """

    if array.dtype == dtype:
        return array
    elif (array.dtype.itemsize == dtype.itemsize and not
            (np.issubdtype(array.dtype, np.number) and
             np.issubdtype(dtype, np.number))):
        # Includes a special case when both dtypes are at least numeric to
        # account for ticket #218: https://trac.assembla.com/pyfits/ticket/218
        return array.view(dtype)
    else:
        return array.astype(dtype)


def _unsigned_zero(dtype):
    """
    Given a numpy dtype, finds its "zero" point, which is exactly in the
    middle of its range.
    """

    assert dtype.kind == 'u'
    return 1 << (dtype.itemsize * 8 - 1)


def _is_pseudo_unsigned(dtype):
    return dtype.kind == 'u' and dtype.itemsize >= 2


def _is_int(val):
    return isinstance(val, (int, long, np.integer))


def _str_to_num(val):
    """Converts a given string to either an int or a float if necessary."""

    try:
        num = int(val)
    except ValueError:
        # If this fails then an exception should be raised anyways
        num = float(val)
    return num


def _pad_length(stringlen):
    """Bytes needed to pad the input stringlen to the next FITS block."""

    return (BLOCK_SIZE - (stringlen % BLOCK_SIZE)) % BLOCK_SIZE


def _normalize_slice(input, naxis):
    """
    Set the slice's start/stop in the regular range.
    """

    def _normalize(indx, npts):
        if indx < -npts:
            indx = 0
        elif indx < 0:
            indx += npts
        elif indx > npts:
            indx = npts
        return indx

    _start = input.start
    if _start is None:
        _start = 0
    elif _is_int(_start):
        _start = _normalize(_start, naxis)
    else:
        raise IndexError('Illegal slice %s; start must be integer.' % input)

    _stop = input.stop
    if _stop is None:
        _stop = naxis
    elif _is_int(_stop):
        _stop = _normalize(_stop, naxis)
    else:
        raise IndexError('Illegal slice %s; stop must be integer.' % input)

    if _stop < _start:
        raise IndexError('Illegal slice %s; stop < start.' % input)

    _step = input.step
    if _step is None:
        _step = 1
    elif _is_int(_step):
        if _step <= 0:
            raise IndexError('Illegal slice %s; step must be positive.'
                             % input)
    else:
        raise IndexError('Illegal slice %s; step must be integer.' % input)

    return slice(_start, _stop, _step)


def _tmp_name(input):
    """
    Create a temporary file name which should not already exist.  Use the
    directory of the input file as the base name of the mkstemp() output.
    """

    if input is not None:
        input = os.path.dirname(input)
    f, fn = tempfile.mkstemp(dir=input)
    os.close(f)
    return fn


if sys.version_info[:2] < (2, 6):
    # In Python 2.5 mmap.mmap is a function that returns an object of type
    # 'mmap.mmap', but the mmap.mmap type is otherwise not accessible through
    # the module
    def _is_mmap(obj):
        return (type(obj).__module__ == 'mmap' and
                type(obj).__name__ == 'mmap')
else:
    def _is_mmap(obj):
        return isinstance(obj, mmap.mmap)


def _get_array_mmap(array):
    """
    If the array has an mmap.mmap at base of its base chain, return the mmap
    object; otherwise return None.
    """

    if _is_mmap(array):
        return array

    base = array
    while hasattr(base, 'base') and base.base is not None:
        if _is_mmap(base.base):
            return base.base
        base = base.base



from __future__ import division # confidence high

import os
import shutil
import stat
import tempfile
import warnings


class PyfitsTestCase(object):
    def setup(self):
        self.data_dir = os.path.join(os.path.dirname(__file__), 'data')
        self.temp_dir = tempfile.mkdtemp(prefix='pyfits-test-')

        # Ignore deprecation warnings--this only affects Python 2.5 and 2.6,
        # since deprecation warnings are ignored by defualt on 2.7
        warnings.resetwarnings()
        warnings.simplefilter('ignore')
        warnings.simplefilter('always', UserWarning)

    def teardown(self):
        shutil.rmtree(self.temp_dir)

    def copy_file(self, filename):
        """Copies a backup of a test data file to the temp dir and sets its
        mode to writeable.
        """

        shutil.copy(self.data(filename), self.temp(filename))
        os.chmod(self.temp(filename), stat.S_IREAD | stat.S_IWRITE)

    def data(self, filename):
        """Returns the path to a test data file."""

        return os.path.join(self.data_dir, filename)

    def temp(self, filename):
        """ Returns the full path to a file in the test temp dir."""

        return os.path.join(self.temp_dir, filename)



from pyfits.hdu.base import register_hdu, unregister_hdu, DELAYED
from pyfits.hdu.compressed import CompImageHDU
from pyfits.hdu.groups import GroupsHDU, GroupData, Group
from pyfits.hdu.hdulist import HDUList
from pyfits.hdu.image import PrimaryHDU, ImageHDU
from pyfits.hdu.streaming import StreamingHDU
from pyfits.hdu.table import TableHDU, BinTableHDU

__all__ = ['HDUList', 'PrimaryHDU', 'ImageHDU', 'TableHDU', 'BinTableHDU',
           'GroupsHDU', 'GroupData', 'Group', 'CompImageHDU', 'StreamingHDU',
           'register_hdu', 'unregister_hdu', 'DELAYED']



import os
import sys
import warnings

try:
    from collections import MutableMapping as __HEADERBASE
except ImportError:
    from UserDict import DictMixin
    class __HEADERBASE(DictMixin, object): # object to make a newstyle class
        pass

from pyfits.card import Card, CardList, RecordValuedKeywordCard, \
                        _ContinueCard, _HierarchCard, create_card, \
                        create_card_from_string, upper_key
from pyfits.util import BLOCK_SIZE, deprecated


PY3K = sys.version_info[:2] >= (3, 0)


class Header(__HEADERBASE):
    """
    FITS header class.

    The purpose of this class is to present the header like a
    dictionary as opposed to a list of cards.

    The attribute `ascard` supplies the header like a list of cards.

    The header class uses the card's keyword as the dictionary key and
    the cards value is the dictionary value.

    The `has_key`, `get`, and `keys` methods are implemented to
    provide the corresponding dictionary functionality.  The header
    may be indexed by keyword value and like a dictionary, the
    associated value will be returned.  When the header contains cards
    with duplicate keywords, only the value of the first card with the
    given keyword will be returned.

    The header may also be indexed by card list index number.  In that
    case, the value of the card at the given index in the card list
    will be returned.

    A delete method has been implemented to allow deletion from the
    header.  When `del` is called, all cards with the given keyword
    are deleted from the header.

    The `Header` class has an associated iterator class `_Header_iter`
    which will allow iteration over the unique keywords in the header
    dictionary.
    """

    # TODO: Allow the header to take a few other types of inputs, for example
    # a list of (key, value) tuples, (key, value, comment) tuples, or a dict
    # of either key: value or key: (value, comment) mappings.  This could all
    # be handled by the underlying CardList I suppose.
    def __init__(self, cards=[], txtfile=None):
        """
        Construct a `Header` from a `CardList` and/or text file.

        Parameters
        ----------
        cards : A list of `Card` objects, optional
            The cards to initialize the header with.

        txtfile : file path, file object or file-like object, optional
            Input ASCII header parameters file.
        """

        # populate the cardlist
        self.ascard = CardList(cards)

        if txtfile:
            # get the cards from the input ASCII file
            self.fromTxtFile(txtfile, not len(self.ascard))
        self._mod = False

    def __len__(self):
        return len(self.ascard)

    def __iter__(self):
        for card in self.ascard:
            yield card.key

    def __contains__(self, key):
        """
        Check for existence of a keyword.

        Parameters
        ----------
        key : str or int
           Keyword name.  If given an index, always returns 0.

        Returns
        -------
        has_key : bool
            Returns `True` if found, otherwise, `False`.
        """

        key = upper_key(key)
        if key[:8] == 'HIERARCH':
            key = key[8:].strip()
        return key in self.ascard
    has_key = deprecated(name='has_key',
                         alternative='`key in header` syntax')(__contains__)

    def __getitem__ (self, key):
        """
        Get a header keyword value.
        """

        card = self.ascard[key]

        if isinstance(card, RecordValuedKeywordCard) and \
           (not isinstance(key, basestring) or '.' not in key):
            return card.strvalue()
        elif isinstance(card, CardList):
            return card
        else:
            return card.value

    def __setitem__ (self, key, value):
        """
        Set a header keyword value.
        """

        card = self.ascard[key]
        self.ascard[key].value = value
        if card._modified:
            self._mod = True

    def __delitem__(self, key):
        """
        Delete card(s) with the name `key`.
        """

        # delete ALL cards with the same keyword name
        if isinstance(key, basestring):
            while True:
                try:
                    del self.ascard[key]
                    self._mod = True
                except:
                    return

        # for integer key only delete once
        else:
            del self.ascard[key]
            self._mod = True

    def __str__(self):
        return str(self.ascard)

    @classmethod
    def fromstring(cls, data):
        """
        Creates an HDU header from a byte string containing the entire header
        data.

        Parameters
        ----------
        data : str
           String containing the entire header.
        """

        if (len(data) % BLOCK_SIZE) != 0:
            raise ValueError('Header size is not multiple of %d: %d'
                             % (BLOCK_SIZE, len(data)))

        cards = []
        keys = []

        # Split the header into individual cards
        for idx in range(0, len(data), Card.length):
            card = create_card_from_string(data[idx:idx + Card.length])
            key = card.key

            if key == 'END':
                break
            else:
                cards.append(card)
                keys.append(key)

        # Deal with CONTINUE cards
        # if a long string has CONTINUE cards, the "Card" is considered
        # to be more than one 80-char "physical" cards.
        idx = len(cards)
        continueimg = []
        for card in reversed(cards):
            idx -= 1
            if idx != 0 and card.key == 'CONTINUE':
                continueimg.append(card._cardimage)
                del cards[idx]
            elif continueimg:
                continueimg.append(card._cardimage)
                continueimg = ''.join(reversed(continueimg))
                cards[idx] = _ContinueCard.fromstring(continueimg)
                continueimg = []

        return cls(CardList(cards, keylist=keys))

    def keys(self):
        """
        Return a list of keys with duplicates removed.

        .. warning::
            There is a surprising incogruity in Header objecets between
            :meth:`Header.keys` and :meth:`Header.iterkeys`.  The latter does
            *not* remove duplicates.  This incongruity exists for historical
            reasons, but is not be design.  In PyFITS 3.1 it is done away with,
            and :meth:Header.keys returns the exact keywords appearin the
            header, including duplicates.
        """

        retval = []

        for key in self.ascard.keys():
            if not key in retval:
                retval.append(key)

        return retval

    def itervalues(self):
        """
        Override itervalues since the default implementation does not
        properly handle duplicate keywords.
        """

        for card in self.ascard:
            yield card.value

    def iteritems(self):
        """
        Override iteritems since the default implementation does not properly
        handle duplicate keywords.
        """

        for card in self.ascard:
            yield (card.key, card.value)

    def values(self):
        """
        Override values since the default implementation does not properly
        handle duplicate keywords.
        """

        return list(self.itervalues())

    def items(self):
        """
        Override items since the default implementation does not propertly
        handle duplicate keywords.
        """

        return list(self.iteritems())

    # Some fixes for compatibility with the Python 3 dict interface, where
    # iteritems -> items, etc.
    if PY3K:  # pragma: py3
        values = itervalues
        items = iteritems
        del itervalues
        del iteritems

    def setdefault(self, keyword, default=''):
        """
        PyFITS < 3.1 won't allow item assignment to keywords that don't already
        exist, but for the setdefault dict method to work at all, it needs to
        be able to add nonexistent keywords with the default value.
        """

        try:
            return self[keyword]
        except KeyError:
            self.update(keyword, default)
            return default

    def update(self, key, value, comment=None, before=None, after=None,
               savecomment=False):
        """
        Update one header card.

        If the keyword already exists, it's value and/or comment will
        be updated.  If it does not exist, a new card will be created
        and it will be placed before or after the specified location.
        If no `before` or `after` is specified, it will be appended at
        the end.

        Parameters
        ----------
        key : str
            keyword

        value : str
            value to be used for updating

        comment : str, optional
            to be used for updating, default=None.

        before : str or int, optional
            name of the keyword, or index of the `Card` before which
            the new card will be placed.  The argument `before` takes
            precedence over `after` if both specified.

        after : str or int, optional
            name of the keyword, or index of the `Card` after which
            the new card will be placed.

        savecomment : bool, optional
            When `True`, preserve the current comment for an existing
            keyword.  The argument `savecomment` takes precedence over
            `comment` if both specified.  If `comment` is not
            specified then the current comment will automatically be
            preserved.
        """

        keylist = RecordValuedKeywordCard.valid_key_value(key, value)

        if keylist:
            keyword = keylist[0] + '.' + keylist[1]
        else:
            keyword = key

        if keyword in self:
            j = self.ascard.index_of(keyword)
            if not savecomment and comment is not None:
                _comment = comment
            else:
                _comment = self.ascard[j].comment
            _card = create_card(key, value, _comment)
            self.ascard[j] = _card
            if before is not None or after is not None:
                self.ascard._pos_insert(_card, before=before, after=after,
                                        replace=True)
        elif before is not None or after is not None:
            _card = create_card(key, value, comment)
            self.ascard._pos_insert(_card, before=before, after=after)
        else:
            self.ascard.append(create_card(key, value, comment))
        if self.ascard._mod:
            self._mod = True

    def copy(self, strip=False):
        """
        Make a copy of the `Header`.

        Parameters
        ----------
        strip : bool, optional
           If True, strip any headers that are specific to one of the standard
           HDU types, so that this header can be used in a different HDU.
        """

        tmp = Header(self.ascard.copy())
        if strip:
            tmp._strip()
        return tmp

    @deprecated(alternative='the ascard attribute')
    def ascardlist(self):
        """
        Returns a `CardList` object.
        """

        return self.ascard

    def rename_key(self, oldkey, newkey, force=False):
        """
        Rename a card's keyword in the header.

        Parameters
        ----------
        oldkey : str or int
            old keyword

        newkey : str
            new keyword

        force : bool
            When `True`, if new key name already exists, force to have
            duplicate name.
        """

        oldkey = upper_key(oldkey)
        newkey = upper_key(newkey)

        if newkey == 'CONTINUE':
            raise ValueError('Can not rename to CONTINUE')

        if newkey in Card._commentary_keys or oldkey in Card._commentary_keys:
            if not (newkey in Card._commentary_keys and
                    oldkey in Card._commentary_keys):
                raise ValueError('Regular and commentary keys can not be '
                                 'renamed to each other.')
        elif (force == 0) and newkey in self:
            raise ValueError('Intended keyword %s already exists in header.'
                             % newkey)

        idx = self.ascard.index_of(oldkey)
        comment = self.ascard[idx].comment
        value = self.ascard[idx].value
        self.ascard[idx] = create_card(newkey, value, comment)

    def add_history(self, value, before=None, after=None):
        """
        Add a ``HISTORY`` card.

        Parameters
        ----------
        value : str
            history text to be added.

        before : str or int, optional
            same as in `Header.update`

        after : str or int, optional
            same as in `Header.update`
        """

        self._add_commentary('history', value, before=before, after=after)

    def add_comment(self, value, before=None, after=None):
        """
        Add a ``COMMENT`` card.

        Parameters
        ----------
        value : str
            text to be added.

        before : str or int, optional
            same as in `Header.update`

        after : str or int, optional
            same as in `Header.update`
        """

        self._add_commentary('comment', value, before=before, after=after)

    def add_blank(self, value='', before=None, after=None):
        """
        Add a blank card.

        Parameters
        ----------
        value : str, optional
            text to be added.

        before : str or int, optional
            same as in `Header.update`

        after : str or int, optional
            same as in `Header.update`
        """

        self._add_commentary(' ', value, before=before, after=after)

    def get_history(self):
        """
        Get all history cards as a list of string texts.
        """

        return [c for c in self.ascard if c.key == 'HISTORY']

    def get_comment(self):
        """
        Get all comment cards as a list of string texts.
        """

        return [c for c in self.ascard if c.key == 'COMMENT']

    def toTxtFile(self, fileobj, clobber=False):
        """
        Output the header parameters to a file in ASCII format.

        Parameters
        ----------
        fileobj : file path, file object or file-like object
            Output header parameters file.

        clobber : bool
            When `True`, overwrite the output file if it exists.
        """

        close_file = False

        # check if the output file already exists
        if isinstance(fileobj, basestring):
            if (os.path.exists(fileobj) and os.path.getsize(fileobj) != 0):
                if clobber:
                    warnings.warn("Overwriting existing file '%s'." % fileobj)
                    os.remove(fileobj)
                else:
                    raise IOError("File '%s' already exist." % fileobj)

            fileobj = open(fileobj, 'w')
            close_file = True

        lines = []   # lines to go out to the header parameters file

        # Add the card image for each card in the header to the lines list

        for j in range(len(self.ascard)):
            lines.append(str(self.ascard[j]) + '\n')

        # Write the header parameter lines out to the ASCII header
        # parameter file
        fileobj.writelines(lines)

        if close_file:
            fileobj.close()

    def fromTxtFile(self, fileobj, replace=False):
        """
        Input the header parameters from an ASCII file.

        The input header cards will be used to update the current
        header.  Therefore, when an input card key matches a card key
        that already exists in the header, that card will be updated
        in place.  Any input cards that do not already exist in the
        header will be added.  Cards will not be deleted from the
        header.

        Parameters
        ----------
        fileobj : file path, file object or file-like object
            Input header parameters file.

        replace : bool, optional
            When `True`, indicates that the entire header should be
            replaced with the contents of the ASCII file instead of
            just updating the current header.
        """

        close_file = False

        if isinstance(fileobj, basestring):
            fileobj = open(fileobj, 'r')
            close_file = True

        lines = fileobj.readlines()

        if close_file:
            fileobj.close()

        if len(self.ascard) > 0 and not replace:
            prevKey = 0
        else:
            if replace:
                self.ascard = CardList([])
            prevKey = 0

        for line in lines:
            card = Card.fromstring(line[:min(80, len(line)-1)])
            card.verify('silentfix')

            if card.key == 'SIMPLE':
                if self.get('EXTENSION'):
                    del self.ascard['EXTENSION']

                self.update(card.key, card.value, card.comment, before=0)
                prevKey = 0
            elif card.key == 'EXTENSION':
                if self.get('SIMPLE'):
                    del self.ascard['SIMPLE']

                self.update(card.key, card.value, card.comment, before=0)
                prevKey = 0
            elif card.key == 'HISTORY':
                if not replace:
                    items = self.items()
                    idx = 0

                    for item in items:
                        if item[0] == card.key and item[1] == card.value:
                            break
                        idx += 1

                    if idx == len(self.ascard):
                        self.add_history(card.value, after=prevKey)
                        prevKey += 1
                else:
                    self.add_history(card.value, after=prevKey)
                    prevKey += 1
            elif card.key == 'COMMENT':
                if not replace:
                    items = self.items()
                    idx = 0

                    for item in items:
                        if item[0] == card.key and item[1] == card.value:
                            break
                        idx += 1

                    if idx == len(self.ascard):
                        self.add_comment(card.value, after=prevKey)
                        prevKey += 1
                else:
                    self.add_comment(card.value, after=prevKey)
                    prevKey += 1
            elif card.key == '        ':
                if not replace:
                    items = self.items()
                    idx = 0

                    for item in items:
                        if item[0] == card.key and item[1] == card.value:
                            break
                        idx += 1

                    if idx == len(self.ascard):
                        self.add_blank(card.value, after=prevKey)
                        prevKey += 1
                else:
                    self.add_blank(card.value, after=prevKey)
                    prevKey += 1
            else:
                if isinstance(card, _HierarchCard):
                    prefix = 'hierarch '
                else:
                    prefix = ''

                self.update(prefix + card.key,
                                     card.value,
                                     card.comment,
                                     after=prevKey)
                prevKey += 1

    def _add_commentary(self, key, value, before=None, after=None):
        """
        Add a commentary card.

        If `before` and `after` are `None`, add to the last occurrence
        of cards of the same name (except blank card).  If there is no
        card (or blank card), append at the end.
        """

        # The maximum value in each card can be the maximum card length minus
        # the key length minus a space (or just minus one if the key is blank)
        maxlen = Card.length - len(key.strip()) - 1

        if len(value) <= maxlen:
            # The value can fit in a single card
            new_cards = [Card(key, value)]
        else:
            # The value must be split across multiple consecutive commentary
            # cards
            idx = 0
            new_cards = []
            while idx < len(value):
                new_cards.append(Card(key, value[idx:idx + maxlen]))
                idx += maxlen

        if before is not None or after is not None:
            if after:
                new_cards = reversed(new_cards)
            for card in new_cards:
                self.ascard._pos_insert(card, before=before, after=after)
        else:
            if key[0] == ' ':
                for card in new_cards:
                    useblanks = card.cardimage != (' ' * 80)
                    self.ascard.append(card, useblanks=useblanks, bottom=True)
            else:
                try:
                    last = self.ascard.index_of(key, backward=True)
                    for card in reversed(new_cards):
                        self.ascard.insert(last + 1, card)
                except:
                    for card in new_cards:
                        self.ascard.append(card, bottom=True)

        self._mod = True

    def _strip(self):
        """
        Strip cards specific to a certain kind of header.

        Strip cards like ``SIMPLE``, ``BITPIX``, etc. so the rest of
        the header can be used to reconstruct another kind of header.
        """

        # TODO: Previously this only deleted some cards specific to an HDU if
        # _hdutype matched that type.  But it seemed simple enough to just
        # delete all desired cards anyways, and just ignore the KeyErrors if
        # they don't exist.
        # However, it might be desirable to make this extendable somehow--have
        # a way for HDU classes to specify some headers that are specific only
        # to that type, and should be removed otherwise.

        try:
            if 'NAXIS' in self:
                naxis = self['NAXIS']
            else:
                naxis = 0

            if 'TFIELDS' in self:
                tfields = self['TFIELDS']
            else:
                tfields = 0

            for idx in range(naxis):
                del self['NAXIS' + str(idx + 1)]

            for name in ('TFORM', 'TSCAL', 'TZERO', 'TNULL', 'TTYPE',
                         'TUNIT', 'TDISP', 'TDIM', 'THEAP', 'TBCOL'):
                for idx in range(tfields):
                    del self[name + str(idx + 1)]

            for name in ('SIMPLE', 'XTENSION', 'BITPIX', 'NAXIS', 'EXTEND',
                         'PCOUNT', 'GCOUNT', 'GROUPS', 'BSCALE', 'BZERO',
                         'TFIELDS'):
                del self[name]
        except KeyError:
            pass
